from .base import *


class Resizer(NormalElement):
    name = "Resize"
    comment = "Resizes its inputs"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [IntParameter("width", value=120), IntParameter("height", value=120)]

    def process_inputs(self, inputs, outputs, parameters):
        outputs["output"] = Data(cv.resize(inputs["input"].value, (parameters["width"], parameters["height"])))


class AutoResizer(NormalElement):
    name = "Automatic resizer"
    comment = "Resizes its inputs to match first input's size"

    def get_attributes(self):
        return [Input("main"), Input("others", multiple=True)], \
               [Output("main"), Output("others", desequencing=True)], \
            []

    def process_inputs(self, inputs, outputs, parameters):
        size = inputs["main"].value.shape
        size = (size[1], size[0])
        outputs["main"] = Data(cv.resize(inputs["main"].value, size))
        outputs["others"] = Data(cv.resize(inputs["others"].value, size))


class Cropper(NormalElement):
    name = "Cropper"
    comment = "Crops the image"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [IntParameter("width", min_=1, max_=10000, value=128),
                IntParameter("height", min_=1, max_=10000, value=128),
                FloatParameter("x", "x", 0.5, 0, 1), FloatParameter("y", "y", 0.5, 0, 1)]

    def process_inputs(self, inputs, outputs, parameters):
        img = inputs["input"].value
        cw = int(img.shape[1] * parameters["x"])
        w2 = parameters["width"] // 2
        ch = int(img.shape[0] * parameters["y"])
        h2 = parameters["height"] // 2
        crop = img[max(0, ch - h2):ch + h2, max(0, cw - w2):cw + w2]
        outputs["output"] = Data(crop)



class AutoCrop(NormalElement):
    name = "Auto cropper"
    comment = "Automatically removes borders from image"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [FloatParameter("thr", "Threshold", 0), ComboboxParameter("thr_type", [("lighter", "lighter"), ("darker", "darker")], "Crop values")]

    def process_inputs(self, inputs, outputs, parameters):
        image = inputs["input"].value
        threshold = parameters["thr"]
        threshold_type = parameters["thr_type"]

        if threshold_type == "lighter":
            if len(image.shape) == 3:
                i = image.min(axis=2)
            else: i = image
            mask = i >= threshold
        elif threshold_type == "darker":
            if len(image.shape) == 3:
                i = image.max(axis=2)
            else: i = image
            mask = i <= threshold
        else:
            raise Exception("Wrong threshold type")

        horizontal = np.min(mask, 0).tolist()
        vertical = np.min(mask, 1).tolist()

        assert isinstance(horizontal, list)
        assert isinstance(vertical, list)

        left = horizontal.index(0)
        right_margin = horizontal[::-1].index(0)
        right = len(horizontal) - right_margin

        top = vertical.index(0)
        bottom_margin = vertical[::-1].index(0)
        bottom = len(vertical) - bottom_margin

        cropped = image[top:bottom, left:right, ...] + 0

        outputs["output"] = ImageData(cropped)


register_elements_auto(__name__, locals(), "Transforms", 5)




from PyQt5 import QtGui, QtCore

from ..diagram.elements import get_sorted_elements
from .elements import *
from .mimedata import *


class Toolbox(StyledWidget):
    help = """\
Element toolbox

Drag & drop element name to add it to the active diagram"""
    filter_help = """Search for the elements by part of the name"""

    def __init__(self):
        super(Toolbox, self).__init__()
        self.setObjectName("Toolbox")
        label = QLabel("Image Processing Toolbox")
        label.setObjectName("ToolboxTitle")
        elements_list = ElementsList()
        elements_list.setObjectName("ToolboxTree")
        self.filter_input = QLineEdit()
        self.filter_input.setPlaceholderText("Search toolbox...")
        self.filter_input.setToolTip(self.filter_help)
        self.filter_input.textChanged.connect(elements_list.filter_changed)

        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(label)
        layout.addWidget(self.filter_input)
        layout.addWidget(elements_list)
        self.setLayout(layout)

        self.setToolTip(self.help)


class BoldElementGroupDelegate(QStyledItemDelegate):
    def paint(self, painter, option, index):
        isIndexParent = not index.parent().isValid()
        if isIndexParent:
            option.font.setWeight(QFont.Bold)
        QStyledItemDelegate.paint(self, painter, option, index)


class ElementsList(QTreeView):
    def __init__(self):
        super(ElementsList, self).__init__()
        self.class_mapper = None
        elements_tree_model = self.prepare_elements_model()
        self.filter_proxy = FilterProxy(self)
        self.filter_proxy.setSourceModel(elements_tree_model)
        self.filter_proxy.setFilterCaseSensitivity(QtCore.Qt.CaseInsensitive)
        self.setModel(self.filter_proxy)
        self.setItemDelegate(BoldElementGroupDelegate(self))
        self.collapseAll()
        self.header().hide()
        self.setFocusPolicy(QtCore.Qt.NoFocus)
        self.last_spawned_element = 0

    def prepare_elements_model(self):
        model = QStandardItemModel()
        nodes = [(elements, QStandardItem(package)) for package, elements in get_sorted_elements()]
        element_types = [node[0] for node in nodes]
        elements_types_list = flatten_list(element_types)
        self.class_mapper = ClassStringMapper(elements_types_list)

        for elements, node in nodes:
            node.setCheckable(False)
            node.setEditable(False)
            for element in elements:
                item = QStandardItem(element.name)
                item.setEditable(False)
                item.setCheckable(False)
                row = [item,
                       QStandardItem(element.comment),
                       QStandardItem(self.class_mapper.to_string(element))]
                item.setToolTip(element.name + "\n-----------------------------------------\n" + element.comment)
                node.appendRow(row)
            model.appendRow(node)
        return model

    @pyqtSlot(str)
    def filter_changed(self, text):
        self.filter_proxy.setFilterFixedString(text)
        if text: self.expandAll()
        else: self.collapseAll()

    def is_draggable_item_selected(self):
        return len(self.selectedIndexes()) > 2

    def mousePressEvent(self, event):
        QTreeView.mousePressEvent(self, event)
        if self.is_draggable_item_selected():
            index = self.selectedIndexes()[2]
            source_selected = index.model().mapToSource(index)
            selected = self.model().sourceModel().itemFromIndex(source_selected)
            if event.button() == QtCore.Qt.LeftButton:
                drag = QDrag(self)
                mime_data = QtCore.QMimeData()
                mime_data.setText(Mime.NEW_ELEMENT)
                selected_class = self.class_mapper.to_class(str(selected.text()))
                self.last_spawned_element = selected_class()
                self.last_spawned_element.label.setMinimumWidth(100)  # workaround for drag problems - with pixmap width < 100 and adjusted HotSpot
                self.last_spawned_element.setParent(self)
                img = self.last_spawned_element.grab()
                self.last_spawned_element.label.setMinimumWidth(0)
                drag.setMimeData(mime_data)
                drag.setPixmap(img)
                drag.setHotSpot(QtCore.QPoint(img.width() / 2, img.height() / 2))
                drag.exec_()
                if self.last_spawned_element.parent() is self:
                    self.last_spawned_element.setParent(None)
                    self.last_spawned_element.deleteLater()
                else:
                    # todo: move this functionality to elements self logic
                    self.last_spawned_element.recalculate(True, True, True)
        elif event.button() == QtCore.Qt.LeftButton:
            self.setExpanded(self.currentIndex(), not self.isExpanded(self.currentIndex()))

        self.clearSelection()


class FilterProxy(QtCore.QSortFilterProxyModel):
    def filterAcceptsRow(self, source_row, source_parent):
        if not self.filterRegExp().isEmpty():
            source_index = self.sourceModel().index(source_row, self.filterKeyColumn(), source_parent)
            if source_index.isValid():
                child_rows_count = self.sourceModel().rowCount(source_index)
                for i in range(child_rows_count):
                    if self.filterAcceptsRow(i, source_index):
                        return True
                key = self.sourceModel().data(source_index, self.filterRole())
                if hasattr(key, "toString"): key = str(key.toString())
                return self.filterRegExp().indexIn(key) >= 0
        return super(FilterProxy, self).filterAcceptsRow(source_row, source_parent)


class ClassStringMapper:
    def __init__(self, class_list):
        self.map = {}
        for class_type in class_list:
            self.map[class_type.__name__] = class_type

    def to_string(self, class_type):
        return class_type.__name__

    def to_class(self, class_string):
        return self.map[class_string]


def flatten_list(list_of_list):
    return [item for sublist in list_of_list for item in sublist]



from .base import *


class ColorNormalizer(NormalElement):
    name = "Color normalizer"
    comment = "Normalizes each color by given mean and standard deviation"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [FloatParameter("mean", value=128, min_=-255, max_=255),
                FloatParameter("std dev", value=64, min_=-255, max_=255)]

    def process_inputs(self, inputs, outputs, parameters):
        image = inputs["input"].value
        outmean = parameters["mean"]
        outstddev = parameters["std dev"]

        mean, stddev = cv.meanStdDev(image)
        mean = mean.reshape(-1)
        stddev = stddev.reshape(-1)
        stddev[stddev == 0] = 1

        output = (image.astype(np.float32) - mean) * (outstddev/stddev) + outmean

        if image.dtype == np.uint8: np.clip(output, 0, 255, output)
        elif image.dtype == np.int8: np.clip(output, -128, 127, output)
        elif image.dtype == np.uint16: np.clip(output, 0, 65535, output)
        elif image.dtype == np.int16: np.clip(output, -32768, 32767, output)

        outputs["output"] = Data(output.astype(image.dtype))


class OpenCVInRange(NormalElement):
    name = "InRange"
    comment = "Inrange"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [
                    IntParameter("min val", "Minimum value", 0, min_=0, max_=255),
                    IntParameter("max val", "Maximum value", 255, min_=0, max_=255)
               ]

    def process_inputs(self, inputs, outputs, parameters):
        image = np.copy(inputs["input"].value)
        min = parameters["min val"]
        max = parameters["max val"]

        image = cv.inRange(image, min, max)
        outputs["output"] = Data(image)


class OpenCVInRange3D(NormalElement):
    name = "InRange3D"
    comment = "Inrange"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [
                    IntParameter("1 min val", "1:Minimum value", 0, min_=0, max_=255),
                    IntParameter("1 max val", "1:Maximum value", 255, min_=0, max_=255),
                    IntParameter("2 min val", "2:Minimum value", 0, min_=0, max_=255),
                    IntParameter("2 max val", "2:Maximum value", 255, min_=0, max_=255),
                    IntParameter("3 min val", "3:Minimum value", 0, min_=0, max_=255),
                    IntParameter("3 max val", "3:Maximum value", 255, min_=0, max_=255)
               ]

    def process_inputs(self, inputs, outputs, parameters):
        image = np.copy(inputs["input"].value)
        min1 = parameters["1 min val"]
        max1 = parameters["1 max val"]
        min2 = parameters["2 min val"]
        max2 = parameters["2 max val"]
        min3 = parameters["3 min val"]
        max3 = parameters["3 max val"]

        image = cv.inRange(image, (min1, min2, min3), (max1, max2, max3))
        outputs["output"] = Data(image)


class ContrastChange(NormalElement):
    name = "Contrast change"
    comment = "Changes contrast of the image"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [FloatParameter("factor")]

    def process_channels(self, inputs, outputs, parameters):
        image = inputs["input"].value

        avg = cv.mean(image)
        output = cv.add(cv.subtract(np.float64(image), avg) * parameters["factor"], avg)
        output = cv.add(np.zeros(output.shape, image.dtype), output, dtype=cvtypes[image.dtype.name])

        outputs["output"] = Data(output)


register_elements_auto(__name__, locals(), "Color", 4)



import numpy as np

from cvlab.diagram.elements.base import *


def round_11(data, middle):
    if middle:
        delta = 0.33
    else:
        delta = 0
    return -1 * (data < delta) + 1 * (data >= delta)


def round_05(data, middle, classes_count=2):
    if middle:
        d1, d2 = 0.34, 0.16
    else:
        d1, d2 = 0.5, 0

    ret = data * 0
    for c in range(classes_count - 1):  # todo: dla duzej liczby klas to bedzie meeega wolne...
        ret += c * np.logical_and(data >= c - d1, data < c + d1)
        if middle:
            c += 0.5
            ret += c * np.logical_and(data >= c - d2, data < c + d2)
    ret += (classes_count - 1) * (data >= classes_count - 1 - d1)
    return ret


class Trainable(NormalElement):
    def train(self, train_data, responses, sample_weights):
        pass

    def predict(self, v):
        pass

    def predict_all(self, vvv):
        return [self.predict(v) for v in vvv]

    # TODO: to wszystko tutaj jest DO BANI. Trzeba przedyskutowac i zmienic, bo jest za duzo mozliwosci aktualnie.
    # jachoo: moja sugestia - przerobic WSZYSTKIE klasyfikatory na wielowyjsciowe (czyli zamiast numerow klas zeby byly
    #         zawsze prawdopodobienstwa. To uprosci BARDZO liczenie wynikow procentowych itp.
    #         sa dwa minusy w takim wypadku: utrata spojnosci z bibliotekami i troche wieksze zuzycie pamieci

    # scoring defs
    TYPE_CLASSES = 0  # klasy numerowane od 0
    TYPE_CLASSES_WITH_MIDDLE = 1  # j.w. oraz odpowiedz "cos pomiedzy" (wartosci 0.5, 1.5 itd.)
    TYPE_REAL_CENTER_0 = 2  # wyniki bedace liczbami rzeczywistymi typowo od -1 do 1
    TYPE_REAL_CENTER_0_WITH_MIDDLE = 3  # j.w. oraz odpowiedz "cos pomiedzy" (wartosci bliskie 0)
    TYPE_REAL_CENTER_05 = 4  # wyniki bedace liczbami rzeczywistymi typowo od 0 do 1
    TYPE_REAL_CENTER_05_WITH_MIDDLE = 5  # j.w. oraz odpowiedz "cos pomiedzy" (wartosci bliskie 0.5)

    # cross-validation defs
    CV_SIMPLE = 0  # zbiory sa budowane po kolei (pierwsze n/k elementow do pierwszego zbioru itd.)
    CV_STEPPED = 1  # zbiory sa budowane co k elementow (rownomierny podzial danych z calego zbioru wejsciowego)

    output_type = TYPE_CLASSES
    classes_count = 2
    cv_type = CV_SIMPLE

    # returns: [avg errors], [valid percent]
    def score_all(self, predicted, real):
        assert len(predicted) == len(real)
        sample_count = len(predicted)
        errors = None
        valid = None
        outputs = 0
        for p, r in zip(predicted, real):
            self.may_interrupt()
            assert np.alen(p) == np.alen(r)
            if not outputs:
                outputs = np.alen(p)
                errors = np.zeros(outputs)
                valid = np.zeros(outputs)
            e, v = self.score(p, r)
            errors += e
            valid += v
        errors /= sample_count
        valid /= sample_count
        valid_total = valid.sum() / outputs
        return errors, valid, valid_total, sample_count

    @staticmethod
    def format_scores(errors, valid, valid_total, sample_count):
        return "errors: " + str(errors) + " valid: " + str(valid) + " total: " + str(valid_total) + " samples: " + str(
            sample_count)

    def score(self, predicted, real):
        error = abs(predicted - real)
        if self.output_type == self.TYPE_CLASSES:
            valid = 1 * (round_05(predicted, False, self.classes_count) == round_05(real, False, self.classes_count))
        elif self.output_type == self.TYPE_CLASSES_WITH_MIDDLE:
            valid = 1 * (round_05(predicted, True, self.classes_count) == round_05(real, True, self.classes_count))
        elif self.output_type == self.TYPE_REAL_CENTER_0:
            valid = 1 * (round_11(predicted, False) == round_11(real, False))
        elif self.output_type == self.TYPE_REAL_CENTER_0_WITH_MIDDLE:
            valid = 1 * (round_11(predicted, True) == round_11(real, True))
        elif self.output_type == self.TYPE_REAL_CENTER_05:
            valid = 1 * (round_05(predicted, False, 2) == round_05(real, False, 2))
        elif self.output_type == self.TYPE_REAL_CENTER_05_WITH_MIDDLE:
            valid = 1 * (round_05(predicted, True, 2) == round_05(real, True, 2))
        else:
            raise Exception("Wrong scoring type")
        return error, valid

    def cross_validate(self, train_data, responses, sample_weights, k):
        if k <= 0: raise ValueError("Parameter k must be positive")
        outputs = np.alen(responses[0])
        errors, valid, valid_total, total_count = np.zeros(outputs), np.zeros(outputs), 0, 0
        samples = train_data.shape[0]
        if k == 1: k = samples
        cv_step = samples / float(k)
        cv_actual = 0
        for i in range(k):
            if self.cv_type == Trainable.CV_SIMPLE:
                testing_ids = list(range(int(cv_actual), min(int(cv_actual + cv_step), samples)))
                cv_actual += cv_step
            else:
                testing_ids = list(range(i, samples, k))
            current_data = np.delete(train_data, testing_ids, axis=0)
            current_resp = np.delete(responses, testing_ids, axis=0)
            current_weights = np.delete(sample_weights, i, axis=0)
            self.may_interrupt()
            self.train(current_data, current_resp, current_weights)
            self.may_interrupt()
            testing_data = train_data[testing_ids]
            testing_pred = self.predict_all(testing_data)
            testing_real = responses[testing_ids]
            self.may_interrupt()
            e, v, vt, cnt = self.score_all(testing_pred, testing_real)
            errors += e
            valid += v
            valid_total += vt
            total_count += cnt
        return errors / k, valid / k, valid_total / k, total_count

    def train_and_evaluate(self, train_data, responses, sample_weights, cv_k=1):
        if not cv_k:
            cv_scores = None
        else:
            cv_scores = self.cross_validate(train_data, responses, sample_weights, cv_k)

        # todo: dorobic liste blednych rozpoznan (?)

        self.train(train_data, responses, sample_weights)
        train_score = self.score_all(self.predict_all(train_data), responses)

        return """Accuracy:
train data:\t{}
cross-validation:\t{}""".format(self.format_scores(*train_score),
                                self.format_scores(*cv_scores) if cv_scores else "---")

    def get_sample_weights(self, responses):
        counter = {}
        for class_ in responses:
            if class_ in counter:
                counter[class_] += 1
            else:
                counter[class_] = 1
        samples = len(responses)
        length = len(counter)
        sample_weights_map = {}
        for class_, count in counter.items():
            weight = float(samples)/(length * count)
            sample_weights_map[class_] = weight
        weights = []
        for class_ in responses:
            weight = sample_weights_map[class_]
            weights.append(weight)
        return np.array(weights, np.float32)



from distutils.util import strtobool

from PyQt5.QtCore import Qt, pyqtSlot
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from . import config
from .styles import StyleManager


class MenuBar(QMenuBar):
    def __init__(self, parent):
        super(MenuBar, self).__init__(parent)
        main_window = parent
        main_window.setMenuBar(self)

        file_menu = self.addMenu('&File')
        file_menu.addAction(NewAction(file_menu, main_window))
        file_menu.addAction(OpenAction(file_menu, main_window))
        file_menu.addAction(SaveAction(file_menu, main_window))
        file_menu.addAction(SaveAsAction(file_menu, main_window))
        file_menu.addAction(CloseDiagramAction(file_menu, main_window))
        file_menu.addSeparator()
        file_menu.addAction(CloseAppAction(file_menu, main_window))

        edit_menu = self.addMenu("&Edit")
        edit_menu.addAction(CenterElementsAction(edit_menu, main_window))
        edit_menu.addAction(SelectAllAction(edit_menu, main_window))
        edit_menu.addAction(DeleteSelectedAction(edit_menu, main_window))

        view_menu = self.addMenu('&View')
        view_menu.addMenu(ColorThemeMenu(view_menu, main_window))
        view_menu.addAction(HighQualityAction(view_menu, main_window))
        view_menu.addAction(LivePreviewsAction(view_menu, main_window))
        view_menu.addAction(PreviewOnTopAction(view_menu, main_window))
        view_menu.addAction(ResetViewAction(view_menu, main_window))
        view_menu.addAction(ExperimentalElementsAction(view_menu, main_window))

        help_menu = self.addMenu("&Help")
        help_menu.addAction(AboutAction(help_menu, main_window))


class Action(QAction):
    def __init__(self, name, parent, main_window):
        super(Action, self).__init__(name, parent)
        self.main_window = main_window
        self.settings = config.ConfigWrapper.get_settings()


class SimpleAction(Action):
    def __init__(self, name, parent, main_window, action=None, shortcut=None, icon=None):
        super().__init__(name, parent, main_window)

        if shortcut:
            if not isinstance(shortcut, QKeySequence):
                shortcut = QKeySequence(shortcut)
            self.setShortcut(shortcut)

        if icon:
            StyleManager.icons.set_icon(self, icon)

        self.action = action
        self.triggered.connect(self.do_action)

    @pyqtSlot()
    def do_action(self):
        self.action(self.main_window)


class WorkareaAction(SimpleAction):
    def __init__(self, name, parent, main_window, action, shortcut=None, icon=None):
        super().__init__(name, parent, main_window, action, shortcut=shortcut, icon=icon)

    @pyqtSlot()
    def do_action(self):
        workarea = self.main_window.diagram_manager.current_workarea()
        if workarea: self.action(workarea)


class NewAction(SimpleAction):
    def __init__(self, parent, main_window):
        super().__init__('&New', parent, main_window, shortcut=Qt.CTRL + Qt.Key_N, icon="doc 6")

    @pyqtSlot()
    def do_action(self):
        self.main_window.diagram_manager.open_diagram()


class OpenAction(SimpleAction):
    def __init__(self, parent, main_window):
        super(OpenAction, self).__init__('&Open...', parent, main_window, shortcut=Qt.CTRL + Qt.Key_O, icon="folder")

    @pyqtSlot()
    def do_action(self):
        self.main_window.diagram_manager.open_diagram_browse()


class SaveAsAction(SimpleAction):
    def __init__(self, parent, main_window):
        super().__init__('&Save as...', parent, main_window, None, Qt.CTRL + Qt.SHIFT + Qt.Key_S, "floppy disk")

    @pyqtSlot()
    def do_action(self):
        self.main_window.diagram_manager.save_diagram_as()


class SaveAction(SimpleAction):
    def __init__(self, parent, main_window):
        super(SaveAction, self).__init__('Save', parent, main_window, None, Qt.CTRL + Qt.Key_S, "floppy disk")

    @pyqtSlot()
    def do_action(self):
        self.main_window.diagram_manager.save_diagram()


class CloseDiagramAction(SimpleAction):
    def __init__(self, parent, main_window):
        super(CloseDiagramAction, self).__init__('Close diagram', parent, main_window, None, Qt.CTRL + Qt.Key_W, "doc 8")

    @pyqtSlot()
    def do_action(self):
        self.main_window.diagram_manager.close_diagram()


class CloseAppAction(SimpleAction):
    def __init__(self, parent, main_window):
        super(CloseAppAction, self).__init__('E&xit', parent, main_window, None, Qt.ALT + Qt.Key_F4, "power button")

    @pyqtSlot()
    def do_action(self):
        self.main_window.diagram_manager.close_all_diagrams()
        self.main_window.application.quit()


class ColorThemeMenu(QMenu):
    def __init__(self, parent, main_window):
        super(ColorThemeMenu, self).__init__("&Set color theme", parent)
        self.main_window = main_window
        self.fill_styles()
        StyleManager.icons.set_icon(self, "settings 8")

    def fill_styles(self):
        styles = self.main_window.style_manager.get_available_stylesheets()
        current = self.main_window.settings.get_with_default(config.VIEW_SECTION, config.STYLE)
        for style in styles:
            action = self.addAction(style)
            action.setCheckable(True)
            if style == current:
                action.setChecked(True)
            else:
                action.setChecked(False)
            action.triggered.connect(self.style_chosen)

    @pyqtSlot()
    def style_chosen(self):
        chosen_style = self.sender()
        self.apply_style(chosen_style.text())

    def apply_style(self, style):
        for action in self.actions():
            if action.text() == style:
                action.setChecked(True)
            else:
                action.setChecked(False)
        self.main_window.style_manager.apply_stylesheet(style)


class HighQualityAction(Action):
    def __init__(self, parent, main_window):
        super(HighQualityAction, self).__init__('&High quality previews', parent, main_window)
        self.setCheckable(True)
        self.value = bool(strtobool(self.settings.get_with_default(config.VIEW_SECTION, config.VIEW_HQ_OPTION)))
        self.setChecked(self.value)
        StyleManager.icons.set_icon(self, "pic")
        self.triggered.connect(self.toggle_quality)

    @pyqtSlot()
    def toggle_quality(self):
        self.value = not self.value
        self.setChecked(self.value)
        self.settings.set(config.VIEW_SECTION, config.VIEW_HQ_OPTION, self.value)


class LivePreviewsAction(Action):
    def __init__(self, parent, main_window):
        super(LivePreviewsAction, self).__init__('&Show live results', parent, main_window)
        self.setCheckable(True)
        self.value = bool(strtobool(self.settings.get_with_default(config.VIEW_SECTION,
                                                                   config.LIVE_IMAGE_PREVIEW_OPTION)))
        self.setChecked(self.value)
        StyleManager.icons.set_icon(self, "play")
        self.triggered.connect(self.switch)

    @pyqtSlot()
    def switch(self):
        self.value = not self.value
        self.setChecked(self.value)
        self.settings.set(config.VIEW_SECTION, config.LIVE_IMAGE_PREVIEW_OPTION, self.value)


class PreviewOnTopAction(Action):
    def __init__(self, parent, main_window):
        super(PreviewOnTopAction, self).__init__('&Preview always on top', parent, main_window)
        self.setCheckable(True)
        self.value = bool(strtobool(self.settings.get_with_default(config.VIEW_SECTION, config.PREVIEW_ON_TOP_OPTION)))
        self.setChecked(self.value)
        StyleManager.icons.set_icon(self, "pin")
        self.triggered.connect(self.switch)

    @pyqtSlot()
    def switch(self):
        self.value = not self.value
        self.setChecked(self.value)
        self.settings.set(config.VIEW_SECTION, config.PREVIEW_ON_TOP_OPTION, self.value)


class CenterElementsAction(WorkareaAction):
    def __init__(self, parent, main_window):
        super().__init__('Center elements', parent, main_window, self.center_elements, icon="arrow 13")

    def center_elements(self, workarea):
        workarea.workarea.center_elements()
        workarea.scroll_to_upperleft()


class SelectAllAction(WorkareaAction):
    def __init__(self, parent, main_window):
        super().__init__('Select all', parent, main_window, self.select_all, Qt.CTRL + Qt.Key_A, "select")

    def select_all(self, workarea):
        workarea.workarea.selection_manager.select_all_elements()


class DeleteSelectedAction(WorkareaAction):
    def __init__(self, parent, main_window):
        super().__init__('Delete selected', parent, main_window, self.delete_selected, Qt.Key_Delete, "delete")

    def delete_selected(self, workarea):
        workarea.workarea.selection_manager.delete_selected()


class ResetViewAction(WorkareaAction):
    def __init__(self, parent, main_window):
        super(ResetViewAction, self).__init__('Reset view', parent, main_window, self.reset_view, Qt.CTRL + Qt.Key_0, "search")

    def reset_view(self, workarea):
        workarea.workarea.zoom(1.0)
        workarea.scroll_to_upperleft()


class ExperimentalElementsAction(Action):
    def __init__(self, parent, main_window):
        super(ExperimentalElementsAction, self).__init__('&Experimental elements', parent, main_window)
        self.setCheckable(True)
        self.value = bool(strtobool(self.settings.get_with_default(config.ELEMENTS_SECTION, config.EXPERIMENTAL_ELEMENTS)))
        self.setChecked(self.value)
        StyleManager.icons.set_icon(self, "veil")
        self.triggered.connect(self.toggle)

    @pyqtSlot()
    def toggle(self):
        self.value = not self.value
        self.setChecked(self.value)
        self.settings.set(config.ELEMENTS_SECTION, config.EXPERIMENTAL_ELEMENTS, self.value)
        QMessageBox.information(self.main_window, "Information", "You must restart CV Lab to enable/disable experimental elements.")


class AboutAction(SimpleAction):
    message = """\
<h1>CV Lab</h1>
<h2>Computer Vision Laboratory</h2>
<h3>A rapid prototyping tool for computer vision algorithms</h3>

<p>
Homepage on GitHub: <a href="https://github.com/cvlab-ai/cvlab">https://github.com/cvlab-ai/cvlab</a><br/>  
PyPI package: <a href="https://pypi.python.org/pypi/cvlab">https://pypi.python.org/pypi/cvlab</a>
</p>

<p>
<b>Authors:</b> Adam Brzeski, Jan Cychnerski<br/>
<b>License:</b> AGPL-3.0+<br/><br/>
<i>Copyright 2013-2019</i>
</p>"""

    def __init__(self, parent, main_window):
        super().__init__("About", parent, main_window, icon="info")

    @pyqtSlot()
    def do_action(self):
        QMessageBox.about(self.main_window, self.main_window.windowTitle(), self.message)



import os
if os.name != 'nt': raise ImportError("Kinect module only work on Windows")

from datetime import datetime, timedelta
from threading import Event

import math
import time

from cvlab.diagram.elements.base import *
from cvlab.diagram.elements.video_io import Camera

class KinectCamera(NormalElement):
    name = "Kinect (OpenNI)"
    comment = "Reads real-time video and depth from physical kinect device"
    device_lock = Camera.device_lock

    def __init__(self):
        super(KinectCamera, self).__init__()
        self.capture = None
        self.actual_parameters = {"width": 0, "height": 0, "fps": 0}
        self.last_frame_time = datetime.now()
        self.play = Event()  # todo: we shall read playing state from some parameter after load
        self.play.set()
        self.recalculate()

    def get_attributes(self):
        return [], \
            [Output("image"), Output("depth")], \
            [IntParameter("width", value=0, min_=0, max_=4096),
             IntParameter("height", value=0, min_=0, max_=4096),
             FloatParameter("fps", value=15, min_=0.1, max_=120),
             ButtonParameter("pause", self.playpause, "Play / Pause")]

    def playpause(self):
        if self.play.is_set():
            self.play.clear()
        else:
            self.play.set()

    def delete(self):
        self.play.set()
        ThreadedElement.delete(self)
        if self.capture is not None and self.capture.isOpened():
            self.capture.release()

    def process(self):
        parameters = {}
        for name, parameter in self.parameters.items():
            parameters[name] = parameter.get()

        if self.actual_parameters != parameters:
            self.last_frame_time = datetime(2000, 1, 1)

        if self.capture is not None and self.capture.isOpened():
            self.capture.release()
        self.may_interrupt()
        with self.device_lock:
            self.capture = cv.VideoCapture(cv.cv.CV_CAP_OPENNI)
            self.capture.set(cv.cv.CV_CAP_OPENNI_IMAGE_GENERATOR_OUTPUT_MODE, cv.cv.CV_CAP_OPENNI_VGA_30HZ)

        if not self.capture or not self.capture.isOpened():
            raise ProcessingError("Can't open the camera device")

        # todo: uncomment this and implement correctly!
        # if self.actual_parameters["width"] != parameters["width"]:
        #     self.capture.set(3, parameters["width"])
        #     self.actual_parameters["width"] = parameters["width"]
        #     self.may_interrupt()
        #
        # if self.actual_parameters["height"] != parameters["height"]:
        #     self.capture.set(4, parameters["height"])
        #     self.actual_parameters["height"] = parameters["height"]
        #     self.may_interrupt()

        if self.actual_parameters["fps"] != parameters["fps"]:
            self.capture.set(5, int(math.ceil(parameters["fps"])))
            self.actual_parameters["fps"] = parameters["fps"]
            self.may_interrupt()

        while True:
            self.may_interrupt()
            now = datetime.now()
            if now - self.last_frame_time < timedelta(seconds=1.0/parameters["fps"]):
                seconds_to_wait = 1.0/parameters["fps"] - (now-self.last_frame_time).total_seconds()
                breaks = int(round(seconds_to_wait*10+1))
                for _ in range(breaks):
                    time.sleep(seconds_to_wait/breaks)
                    self.may_interrupt()
            self.may_interrupt()
            self.set_state(Element.STATE_BUSY)
            _, image = self.capture.retrieve(0, cv.cv.CV_CAP_OPENNI_BGR_IMAGE)
            _, depth = self.capture.retrieve(0, cv.cv.CV_CAP_OPENNI_DEPTH_MAP)
            self.last_frame_time = datetime.now()
            self.may_interrupt()
            if image is not None and len(image) > 0:
                if parameters["width"] and parameters["height"] and (image.shape[0] != parameters["height"] or image.shape[1] != parameters["width"]):
                    image = cv.resize(image, (parameters["width"], parameters["height"]))
                self.outputs["image"].put(Data(image))

                if parameters["width"] and parameters["height"] and (depth.shape[0] != parameters["height"] or depth.shape[1] != parameters["width"]):
                    depth = cv.resize(depth, (parameters["width"], parameters["height"]))
                self.outputs["depth"].put(Data(depth))

                self.set_state(Element.STATE_READY)
                self.notify_state_changed()
            else:
                return 
            self.may_interrupt()
            self.play.wait()


register_elements_auto(__name__, locals(), "Kinect", 10)





import os
import threading
from collections import OrderedDict

import numpy as np
from PyQt5.QtCore import pyqtSignal, QObject

from . import id_manager


class Parameter(QObject):
    value_changed = pyqtSignal()

    def __init__(self, id, name=None, value=None, *args, **kwargs):
        super(Parameter, self).__init__()
        self.id = id
        self.object_id = id_manager.next_id(self)
        if name is None: name = id
        self.name = name
        self.value = value
        self.lock = threading.RLock()
        self.children = set()
        self.value_changed.emit()

    def connect_child(self, child):
        assert isinstance(child, Parameter)
        if child not in self.children:
            self.children.add(child)
            child.set(self.value)
            child.connect_child(self)

    def disconnect_child(self, child):
        if child in self.children:
            self.children.remove(child)
            child.disconnect_child(self)

    def disconnect_all_children(self):
        for child in frozenset(self.children):
            self.disconnect_child(child)

    def set(self, value):
        with self.lock:
            if self.value == value: return
            self.value = value
        self.value_changed.emit()
        for child in self.children:
            child.set(value)

    def get(self):
        with self.lock:
            return self.value

    def to_json(self):
        return self.value

    def from_json(self, data):
        self.set(data)


class CommentParameter(Parameter):

    status_changed = pyqtSignal()

    def __init__(self, id, name=None, value=""):
        super(CommentParameter, self).__init__(id, name, value)
        self.status_changed.emit()

    def set(self, value):
        Parameter.set(self, str(value))
        self.status_changed.emit()

    def to_json(self):
        return self.value

    def from_json(self, data):
        self.set(data)
        self.status_changed.emit()


class PathParameter(Parameter):
    base_path = None

    def __init__(self, id, name=None, value="", save_mode=False, extension_filter=None, *args, **kwargs):
        super(PathParameter, self).__init__(id, name, value, *args, **kwargs)
        self.save_mode = save_mode
        self.extension_filter = extension_filter  # filter in form "FORMAT (*.ext1 *.ext2 ...)", eg. "IMG (*.jpg *.png)"

    def to_json(self):
        return self.to_json_relative(self.value)

    def from_json(self, data):
        self.set(self.from_json_relative(data))

    @classmethod
    def to_json_relative(cls, path):
        abs_path = os.path.abspath(path).replace("\\","/")
        base_path = os.path.abspath(cls.base_path).replace("\\","/")
        up_base_path = os.path.dirname(base_path)
        if abs_path.startswith(base_path):  # relative path
            relative_path = abs_path[len(base_path):]
            if relative_path.startswith("/"): relative_path = relative_path[1:]
            return relative_path
        elif abs_path.startswith(up_base_path):  # relative to parent directory
            relative_path = abs_path[len(up_base_path):]
            if relative_path.startswith("/"): relative_path = relative_path[1:]
            relative_path = "../" + relative_path
            return relative_path
        else:  # absolute path
            return abs_path

    @classmethod
    def from_json_relative(cls, path):
        path = path.replace("\\","/")
        if os.path.isabs(path):  # absolute path
            return path
        else:  # relative path
            p = os.path.abspath(cls.base_path + "/" + path).replace("\\","/")
            return p


class SavePathParameter(PathParameter):
    def __init__(self, *args, **kwargs):
        super(SavePathParameter, self).__init__(save_mode=True, *args, **kwargs)


class MultiPathParameter(Parameter):
    def to_json(self):
        return [PathParameter.to_json_relative(path) for path in self.value]

    def from_json(self, data):
        self.set([PathParameter.from_json_relative(path) for path in data])


class DirectoryParameter(PathParameter):
    pass


class TextParameter(Parameter):
    def __init__(self, id, name=None, value="", window_title="Text parameter editor", window_content="", live=True):
        super(TextParameter, self).__init__(id, name, value)
        self.window_title = window_title
        self.window_content = window_content
        self.live = live


class NumberParameter(Parameter):
    def __init__(self, id, name=None, value=None, min_=-100000, max_=100000, step=1):
        if value is None:
            if min_ >= 0:
                value = min_
            else:
                value = min(max_, 0)
        super(NumberParameter, self).__init__(id, name, value)
        self.min = min_
        self.max = max_
        self.step = step


class IntParameter(NumberParameter):
    def set(self, value):
        NumberParameter.set(self, int(value))


class FloatParameter(NumberParameter):
    def __init__(self, id, name=None, value=None, min_=-1000, max_=1000, step=0.1):
        super(FloatParameter, self).__init__(id, name, value=value, min_=min_, max_=max_, step=step)

    def set(self, value):
        NumberParameter.set(self, float(value))


class SizeParameter(Parameter):
    def __init__(self, id, name=None, value=(1, 1)):
        super(SizeParameter, self).__init__(id, name, value)
        self.min = 0
        self.max = 10**9

    def set(self, value):
        Parameter.set(self, tuple(value))


class PointParameter(Parameter):
    def __init__(self, id, name=None, value=(0, 0)):
        super(PointParameter, self).__init__(id, name, value)
        self.min = -1
        self.max = 10**9

    def set(self, value):
        Parameter.set(self, tuple(value))


class TwoFloatsParameter(Parameter):
    def __init__(self, id, name=None, value=(0., 0.), min_=-1000, max_=1000):
        super().__init__(id, name, value)
        self.min = min_
        self.max = max_

    def set(self, value):
        Parameter.set(self, tuple(value))


class ScalarParameter(Parameter):
    def __init__(self, id, name=None, value=(0, 0, 0, 0), min_=-1000, max_=1000):
        super(ScalarParameter, self).__init__(id, name, value)
        self.min = min_
        self.max = max_

    def set(self, value):
        value = list(value) + [0,0,0,0]
        value = tuple(value[:4])
        Parameter.set(self, value)


class ComboboxParameter(Parameter):
    def __init__(self, id, values, name=None, default_value_idx=None):
        super(ComboboxParameter, self).__init__(id, name)
        self.values = OrderedDict(values)
        if default_value_idx is not None:
            self.set(list(self.values.values())[default_value_idx])
        else:
            self.value = list(self.values.values())[0]

    def to_json(self):
        # return list(self.values.values()).index(self.value)  # cv-lab v1.0
        index = list(self.values.values()).index(self.value)
        name = list(self.values)[index]
        return {"name": name, "value": self.value}

    def from_json(self, data):
        if isinstance(data, int):  # cv-lab v1.0
            if 0 <= data < len(self.values):
                value = list(self.values.values())[data]
            else:
                print("Warning: Cannot decode parameter '{}' value '{}'".format(self.name, data))
                value = data
        else: # cv-lab v1.1+
            value = data["value"]

        self.set(value)


class ButtonParameter(Parameter):
    def __init__(self, id, callback, name=None):
        super(ButtonParameter, self).__init__(id, name)
        self.callback = callback

    def to_json(self):
        return ""

    def from_json(self, data):
        pass

    def clicked(self):
        self.callback()


class MatrixParameter(Parameter):
    def to_json(self):
        return self.value.tolist()

    def from_json(self, data):
        self.set(np.array(data))



import json

from .parameters import PathParameter


class ComplexJsonEncoder(json.JSONEncoder):
    def __init__(self, base_path=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.base_path = base_path

    def encode(self, o):
        PathParameter.base_path = self.base_path
        encoded = super().encode(o)
        PathParameter.base_path = None
        return encoded

    def default(self, o):
        if hasattr(o, 'to_json'):
                return o.to_json()
        else:
            raise TypeError('Object of type %s with value of %s is not JSON serializable' % (type(o), repr(o)))


class ComplexJsonDecoder(json.JSONDecoder):
    def __init__(self, diagram, base_path):
        json.JSONDecoder.__init__(self, object_hook=self.dict_to_object)
        self.diagram_instance = diagram
        self.base_path = base_path

    def decode(self, s):
        PathParameter.base_path = self.base_path
        super().decode(s)
        PathParameter.base_path = None

    def dict_to_object(self, d):
        if "_type" in d and d["_type"] == "element":
            class_name = d["class"]
            module_name = d["module"]
            from .elements import get_element
            element = get_element(module_name + "." + class_name)()
            element.from_json(d)
            return element
        if "_type" in d and d["_type"] == "diagram":
            self.diagram_instance.from_json(d)
            return self.diagram_instance
        return d



import math

from .base import *


class InPaint(NormalElement):
    name = "Inpaint"
    comment = "Inpaint"
    package = "Photo"

    def get_attributes(self):
        return [Input("input"), Input("mask")], \
               [Output("output")], \
               [IntParameter("radius", value=3, min_=1, max_=100), IntParameter("flags")]

    def process_channels(self, inputs, outputs, parameters):
        output = cv.inpaint(inputs["input"].value, inputs["mask"].value, parameters["radius"], parameters["flags"])
        outputs["output"] = Data(output)


class GetGaborKernel(InputElement):
    name = "getGaborKernel"
    comment = "getGaborKernel"
    package = "Data generation"

    def get_attributes(self):
        return [], [Output("output")], \
               [
                    IntParameter("ksize", "ksize", 1, min_=1, max_=255, step=2),
                    FloatParameter("sigma", "sigma", 0.0, min_=0.0, max_=100.0),
                    FloatParameter("theta", "theta", 0.0, min_=0.0, max_=math.pi, step=0.01),
                    FloatParameter("lambda", "lambda", 0.0, min_=0.0, max_=100.0),
                    FloatParameter("gamma", "gamma", math.pi * 0.5, min_=0.0, max_=100.0, step=0.1),
                    FloatParameter("psi", "psi", 0.0, min_=0.0, max_=100.0),
                    ComboboxParameter("ktype", values={
                        "CV_64F": cv.CV_64F,
                        "CV_32F": cv.CV_32F
                    })
               ]

    def process_inputs(self, inputs, outputs, parameters):
        ksize = parameters["ksize"]
        sigma = parameters["sigma"]
        theta = parameters["theta"]
        lambda_ = parameters["lambda"]
        gamma = parameters["gamma"]
        psi = parameters["psi"]
        ktype = parameters["ktype"]

        image = cv.getGaborKernel(ksize=(ksize, ksize), sigma=sigma, theta=theta, lambd=lambda_, gamma=gamma, psi=psi, ktype=ktype)
        outputs["output"] = Data(image)


register_elements_auto(__name__, locals(), "Image generation", 7)



import os
import re
from glob import glob

from PyQt5.QtWidgets import QAction, QMenu

from cvlab.diagram.elements import add_plugin_callback


def get_menu(main_window, title):
    titles = title.split("/")

    menu = main_window.menuBar()

    simple = lambda text: re.sub(r"[^0-9a-zA-Z]+", "", text)

    for title in titles:
        for child in menu.findChildren(QMenu):
            if simple(child.title()) == simple(title):
                menu = child
                break
        else:
            menu = menu.addMenu(title)

    return menu


class OpenExampleAction(QAction):
    def __init__(self, parent, path):
        super().__init__(parent)
        name = os.path.basename(path).replace(".cvlab", "").title()
        self.setText(name)
        self.path = path
        self.triggered.connect(self.open)

    def open(self):
        self.parent().diagram_manager.open_diagram_from_path(self.path)


def add_samples_submenu_callback(main_window, submenu_name, samples_directory):
    menu_title = 'E&xamples/'+submenu_name
    samples = glob(samples_directory + "/*.cvlab")
    samples.sort()

    print("Adding {} sample diagrams to '{}'".format(len(samples), menu_title))

    menu = get_menu(main_window, menu_title)

    for sample in samples:
        menu.addAction(OpenExampleAction(main_window, sample))


def add_samples_submenu(submenu_name, samples_directory):
    callback = lambda main_window: add_samples_submenu_callback(main_window, submenu_name, samples_directory)
    add_plugin_callback(callback)


add_samples_submenu('Basics', os.path.dirname(__file__))



import json
import os
from datetime import datetime, timedelta

import threading
from os import path
from threading import Lock, Event, Thread
import numpy as np
import cv2 as cv

from PyQt5.QtCore import Qt, QObject, pyqtSlot, pyqtSignal, QSize, QTimer
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *


class Keys:
    NONE = -1
    MOUSE = -2

    NUM0 = 48
    NUM1 = 49
    NUM2 = 50
    NUM3 = 51
    NUM4 = 52
    NUM5 = 53
    NUM6 = 54
    NUM7 = 55
    NUM8 = 56
    NUM9 = 57
    A = 65
    Aacute = 193
    Acircumflex = 194
    acute = 180
    AddFavorite = 16777408
    Adiaeresis = 196
    AE = 198
    Agrave = 192
    Alt = 16777251
    AltGr = 16781571
    Ampersand = 38
    Any = 32
    Apostrophe = 39
    ApplicationLeft = 16777415
    ApplicationRight = 16777416
    Aring = 197
    AsciiCircum = 94
    AsciiTilde = 126
    Asterisk = 42
    At = 64
    Atilde = 195
    AudioCycleTrack = 16777478
    AudioForward = 16777474
    AudioRandomPlay = 16777476
    AudioRepeat = 16777475
    AudioRewind = 16777413
    Away = 16777464
    B = 66
    Back = 16777313
    BackForward = 16777414
    Backslash = 92
    Backspace = 16777219
    Backtab = 16777218
    Bar = 124
    BassBoost = 16777331
    BassDown = 16777333
    BassUp = 16777332
    Battery = 16777470
    Bluetooth = 16777471
    Book = 16777417
    BraceLeft = 123
    BraceRight = 125
    BracketLeft = 91
    BracketRight = 93
    BrightnessAdjust = 16777410
    brokenbar = 166
    C = 67
    Calculator = 16777419
    Calendar = 16777444
    Call = 17825796
    Camera = 17825824
    CameraFocus = 17825825
    Cancel = 16908289
    CapsLock = 16777252
    Ccedilla = 199
    CD = 16777418
    cedilla = 184
    cent = 162
    Clear = 16777227
    ClearGrab = 16777421
    Close = 16777422
    Codeinput = 16781623
    Colon = 58
    Comma = 44
    Community = 16777412
    Context1 = 17825792
    Context2 = 17825793
    Context3 = 17825794
    Context4 = 17825795
    ContrastAdjust = 16777485
    Control = 16777249
    Copy = 16777423
    copyright = 169
    currency = 164
    Cut = 16777424
    D = 68
    Dead_Abovedot = 16781910
    Dead_Abovering = 16781912
    Dead_Acute = 16781905
    Dead_Belowdot = 16781920
    Dead_Breve = 16781909
    Dead_Caron = 16781914
    Dead_Cedilla = 16781915
    Dead_Circumflex = 16781906
    Dead_Diaeresis = 16781911
    Dead_Doubleacute = 16781913
    Dead_Grave = 16781904
    Dead_Hook = 16781921
    Dead_Horn = 16781922
    Dead_Iota = 16781917
    Dead_Macron = 16781908
    Dead_Ogonek = 16781916
    Dead_Semivoiced_Sound = 16781919
    Dead_Tilde = 16781907
    Dead_Voiced_Sound = 16781918
    degree = 176
    Delete = 16777223
    diaeresis = 168
    Direction_L = 16777305
    Direction_R = 16777312
    Display = 16777425
    division = 247
    Documents = 16777427
    Dollar = 36
    DOS = 16777426
    Down = 16777237
    E = 69
    Eacute = 201
    Ecircumflex = 202
    Ediaeresis = 203
    Egrave = 200
    Eisu_Shift = 16781615
    Eisu_toggle = 16781616
    Eject = 16777401
    End = 16777233
    Enter = 16777221
    Equal = 61
    Escape = 16777216
    ETH = 208
    Excel = 16777428
    Exclam = 33
    exclamdown = 161
    Execute = 16908291
    Explorer = 16777429
    F = 70
    F1 = 16777264
    F10 = 16777273
    F11 = 16777274
    F12 = 16777275
    F13 = 16777276
    F14 = 16777277
    F15 = 16777278
    F16 = 16777279
    F17 = 16777280
    F18 = 16777281
    F19 = 16777282
    F2 = 16777265
    F20 = 16777283
    F21 = 16777284
    F22 = 16777285
    F23 = 16777286
    F24 = 16777287
    F25 = 16777288
    F26 = 16777289
    F27 = 16777290
    F28 = 16777291
    F29 = 16777292
    F3 = 16777266
    F30 = 16777293
    F31 = 16777294
    F32 = 16777295
    F33 = 16777296
    F34 = 16777297
    F35 = 16777298
    F4 = 16777267
    F5 = 16777268
    F6 = 16777269
    F7 = 16777270
    F8 = 16777271
    F9 = 16777272
    Favorites = 16777361
    Finance = 16777411
    Flip = 17825798
    Forward = 16777314
    G = 71
    Game = 16777430
    Go = 16777431
    Greater = 62
    guillemotleft = 171
    guillemotright = 187
    H = 72
    Hangul = 16781617
    Hangul_Banja = 16781625
    Hangul_End = 16781619
    Hangul_Hanja = 16781620
    Hangul_Jamo = 16781621
    Hangul_Jeonja = 16781624
    Hangul_PostHanja = 16781627
    Hangul_PreHanja = 16781626
    Hangul_Romaja = 16781622
    Hangul_Special = 16781631
    Hangul_Start = 16781618
    Hangup = 17825797
    Hankaku = 16781609
    Help = 16777304
    Henkan = 16781603
    Hibernate = 16777480
    Hiragana = 16781605
    Hiragana_Katakana = 16781607
    History = 16777407
    Home = 16777232
    HomePage = 16777360
    HotLinks = 16777409
    Hyper_L = 16777302
    Hyper_R = 16777303
    hyphen = 173
    I = 73
    Iacute = 205
    Icircumflex = 206
    Idiaeresis = 207
    Igrave = 204
    Insert = 16777222
    iTouch = 16777432
    J = 74
    K = 75
    Kana_Lock = 16781613
    Kana_Shift = 16781614
    Kanji = 16781601
    Katakana = 16781606
    KeyboardBrightnessDown = 16777398
    KeyboardBrightnessUp = 16777397
    KeyboardLightOnOff = 16777396
    L = 76
    LastNumberRedial = 17825801
    Launch0 = 16777378
    Launch1 = 16777379
    Launch2 = 16777380
    Launch3 = 16777381
    Launch4 = 16777382
    Launch5 = 16777383
    Launch6 = 16777384
    Launch7 = 16777385
    Launch8 = 16777386
    Launch9 = 16777387
    LaunchA = 16777388
    LaunchB = 16777389
    LaunchC = 16777390
    LaunchD = 16777391
    LaunchE = 16777392
    LaunchF = 16777393
    LaunchG = 16777486
    LaunchH = 16777487
    LaunchMail = 16777376
    LaunchMedia = 16777377
    Left = 16777234
    Less = 60
    LightBulb = 16777405
    LogOff = 16777433
    M = 77
    macron = 175
    MailForward = 16777467
    Market = 16777434
    masculine = 186
    Massyo = 16781612
    MediaLast = 16842751
    MediaNext = 16777347
    MediaPause = 16777349
    MediaPlay = 16777344
    MediaPrevious = 16777346
    MediaRecord = 16777348
    MediaStop = 16777345
    MediaTogglePlayPause = 16777350
    Meeting = 16777435
    Memo = 16777404
    Menu = 16777301
    MenuKB = 16777436
    MenuPB = 16777437
    Messenger = 16777465
    Meta = 16777250
    Minus = 45
    Mode_switch = 16781694
    MonBrightnessDown = 16777395
    MonBrightnessUp = 16777394
    mu = 181
    Muhenkan = 16781602
    MultipleCandidate = 16781629
    multiply = 215
    Multi_key = 16781600
    Music = 16777469
    MySites = 16777438
    N = 78
    News = 16777439
    No = 16842754
    nobreakspace = 160
    notsign = 172
    Ntilde = 209
    NumberSign = 35
    NumLock = 16777253
    O = 79
    Oacute = 211
    Ocircumflex = 212
    Odiaeresis = 214
    OfficeHome = 16777440
    Ograve = 210
    onehalf = 189
    onequarter = 188
    onesuperior = 185
    Ooblique = 216
    OpenUrl = 16777364
    Option = 16777441
    ordfeminine = 170
    Otilde = 213
    P = 80
    PageDown = 16777239
    PageUp = 16777238
    paragraph = 182
    ParenLeft = 40
    ParenRight = 41
    Paste = 16777442
    Pause = 16777224
    Percent = 37
    Period = 46
    periodcentered = 183
    Phone = 16777443
    Pictures = 16777468
    Play = 16908293
    Plus = 43
    plusminus = 177
    PowerDown = 16777483
    PowerOff = 16777399
    PreviousCandidate = 16781630
    Print = 16777225
    Printer = 16908290
    Q = 81
    Question = 63
    questiondown = 191
    QuoteDbl = 34
    QuoteLeft = 96
    R = 82
    Refresh = 16777316
    registered = 174
    Reload = 16777446
    Reply = 16777445
    Return = 16777220
    Right = 16777236
    Romaji = 16781604
    RotateWindows = 16777447
    RotationKB = 16777449
    RotationPB = 16777448
    S = 83
    Save = 16777450
    ScreenSaver = 16777402
    ScrollLock = 16777254
    Search = 16777362
    section = 167
    Select = 16842752
    Semicolon = 59
    Send = 16777451
    Shift = 16777248
    Shop = 16777406
    SingleCandidate = 16781628
    Slash = 47
    Sleep = 16908292
    Space = 32
    Spell = 16777452
    SplitScreen = 16777453
    ssharp = 223
    Standby = 16777363
    sterling = 163
    Stop = 16777315
    Subtitle = 16777477
    Super_L = 16777299
    Super_R = 16777300
    Support = 16777454
    Suspend = 16777484
    SysReq = 16777226
    T = 84
    Tab = 16777217
    TaskPane = 16777455
    Terminal = 16777456
    THORN = 222
    threequarters = 190
    threesuperior = 179
    Time = 16777479
    ToDoList = 16777420
    ToggleCallHangup = 17825799
    Tools = 16777457
    TopMenu = 16777482
    Touroku = 16781611
    Travel = 16777458
    TrebleDown = 16777335
    TrebleUp = 16777334
    twosuperior = 178
    U = 85
    Uacute = 218
    Ucircumflex = 219
    Udiaeresis = 220
    Ugrave = 217
    Underscore = 95
    unknown = 33554431
    Up = 16777235
    UWB = 16777473
    V = 86
    Video = 16777459
    View = 16777481
    VoiceDial = 17825800
    VolumeDown = 16777328
    VolumeMute = 16777329
    VolumeUp = 16777330
    W = 87
    WakeUp = 16777400
    WebCam = 16777466
    WLAN = 16777472
    Word = 16777460
    WWW = 16777403
    X = 88
    Xfer = 16777461
    Y = 89
    Yacute = 221
    ydiaeresis = 255
    yen = 165
    Yes = 16842753
    Z = 90
    Zenkaku = 16781608
    Zenkaku_Hankaku = 16781610
    Zoom = 16908294
    ZoomIn = 16777462
    ZoomOut = 16777463


KEY_NONE = Keys.NONE
KEY_MOUSE = Keys.MOUSE


def array_to_pixmap(arr):
    arr = np.array(arr)

    if arr.dtype in (np.float, np.float16, np.float32, np.float64):
        arr = arr * 255
    elif arr.dtype in (np.int16, np.uint16):
        arr = arr // 256

    arr = arr.clip(0, 255).astype(np.uint8)

    if len(arr.shape) == 2 or (len(arr.shape) == 3 and arr.shape[2] == 1):
        arr = cv.cvtColor(arr, cv.COLOR_GRAY2BGRA)
    elif len(arr.shape) == 3 and arr.shape[2] == 3:
        arr = cv.cvtColor(arr, cv.COLOR_BGR2BGRA)

    image = QImage(arr.data, arr.shape[1], arr.shape[0], QImage.Format_ARGB32)
    return QPixmap.fromImage(image)


class PreviewScrollArea(QScrollArea):
    def __init__(self):
        QScrollArea.__init__(self)
        self.lastMouse = None

    def wheelEvent(self, event):
        assert isinstance(event, QWheelEvent)
        event.ignore()

    def mouseMoveEvent(self, event):
        assert isinstance(event, QMouseEvent)
        if self.lastMouse and self.lastMouse != event.pos():
            delta = event.pos() - self.lastMouse
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() - delta.x())
            self.verticalScrollBar().setValue(self.verticalScrollBar().value() - delta.y())
        self.lastMouse = event.pos()
        event.accept()

    def mouseReleaseEvent(self, event):
        self.lastMouse = None
        event.accept()


class PreviewWindow(QFrame):
    minsize = (32, 32)
    maxsize = None
    last_save_dir = ""
    raise_window = False

    key_signal = pyqtSignal(int, int, int)
    move_signal = pyqtSignal()

    def __init__(self, manager, name, image=None, message=None, position=None, size=None, high_quality=False):
        super(PreviewWindow, self).__init__()
        self.setObjectName("Preview window {}".format(name))
        self.setWindowTitle(name)

        self.manager = manager

        desktop = QApplication.instance().desktop()
        if self.maxsize:
            self.maxsize = QSize(*self.maxsize)
        else:
            self.maxsize = desktop.screenGeometry(desktop.screenNumber(self)).size() * 0.95

        self.setMinimumSize(*self.minsize)
        self.setMaximumSize(self.maxsize)

        self.image = None
        self.original = None
        self.message = message
        self.scale = 1.
        self.rotation = 0
        self.quality = Qt.SmoothTransformation if high_quality else Qt.FastTransformation
        self.fixed_size = size

        self.scrollarea = PreviewScrollArea()
        self.scrollarea.setFrameStyle(0)
        self.scrollarea.setFocusPolicy(Qt.NoFocus)

        layout = QGridLayout()
        layout.setContentsMargins(0,0,0,0)
        self.setLayout(layout)
        layout.addWidget(self.scrollarea, 0, 0)

        self.preview = QLabel()
        self.preview.setMouseTracking(False)
        self.preview.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.scrollarea.setWidget(self.preview)

        self.message_label = QLabel(" ")
        self.layout().addWidget(self.message_label, 0, 0, Qt.AlignTop)
        self.message_label.setStyleSheet("QLabel {color:black;background:rgba(255,255,255,32)}")
        self.message_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.message_label.setText("")
        shadow = QGraphicsDropShadowEffect()
        shadow.setBlurRadius(4)
        shadow.setColor(Qt.white)
        shadow.setOffset(0,0)
        self.message_label.setGraphicsEffect(shadow)

        self.blink_widget = QWidget()
        self.blink_widget.hide()
        self.blink_widget.setStyleSheet("border:3px solid red")
        self.blink_timer = QTimer()
        self.blink_timer.setInterval(1000)
        self.blink_timer.timeout.connect(self.blink_)
        layout.addWidget(self.blink_widget, 0, 0)

        self.setImage(image, show=False)

        if image is not None and not size:
            size = self.autoSize()

        if size:
            self.resize(*size)

        if position == 'cursor':
            position = (QCursor.pos().x() - self.size().width()//2, QCursor.pos().y() - self.size().height()//2)

        if position:
            self.move(*position)

        self.showNormal()

    def setImage(self, image, show=True, scale=None, blink=False):
        if image is None: return
        self.original = image
        if isinstance(image, QImage):
            image = QPixmap.fromImage(image)
        elif isinstance(image, QPixmap):
            pass
        else:
            image = array_to_pixmap(image)
        self.image = image
        if not scale:
            scale = self.scale
            if image.width()*scale > self.maxsize.width():
                scale = self.maxsize.width() / image.width()
            if image.height()*scale > self.maxsize.height():
                scale = self.maxsize.height() / image.height()
        self.setZoom(scale)
        if self.message is not None:
            self.message_label.setText(self.message)
        if blink:
            self.blink(True)
        if show:
            self.show()
            if self.raise_window:
                self.raise_()

    def closeEvent(self, event):
        self.scale = 1.0
        self.fixed_size = None

    def setImageAndParams(self, image, show=True, scale=None, position=None, size=None, hq=None, message=None, blink=False):
        if size:
            self.fixed_size = size
        if position:
            self.move(*position)
        if hq is not None:
            if hq: self.quality = Qt.SmoothTransformation
            else: self.quality = Qt.FastTransformation
        if message is not None:
            self.message = message
        self.setImage(image, show=show, scale=scale, blink=blink)

    def setZoom(self, scale):
        self.setParams(scale=scale)

    def setRotation(self, rotation):
        self.setParams(rotation=rotation)

    def setParams(self, scale=None, rotation=None):
        assert isinstance(self.image, QPixmap)

        if scale is None: scale = self.scale
        if rotation is None: rotation = self.rotation

        if scale != 1.0 or rotation:
            transform = QTransform().rotate(rotation).scale(scale,scale)
            pixmap = self.image.transformed(transform, self.quality)
        else:
            pixmap = self.image

        w = pixmap.width()
        h = pixmap.height()

        self.scale = scale
        self.rotation = rotation
        self.preview.setPixmap(pixmap)
        self.preview.setFixedSize(pixmap.size())

        if not self.fixed_size:
            self.resize(w, h)

    def autoSize(self):
        size = self.image.size()
        w = size.width()
        h = size.height()
        return w, h

    def wheelEvent(self, event):
        assert isinstance(event, QWheelEvent)
        event.accept()

        if event.angleDelta().y() > 0:
            s = 1.1
        else:
            s = 1 / 1.1
        self.setZoom(s * self.scale)

        scrollX = self.scrollarea.horizontalScrollBar().value()
        posX = event.x()
        newX = s * (scrollX + posX) - posX
        self.scrollarea.horizontalScrollBar().setValue(int(newX))

        scrollY = self.scrollarea.verticalScrollBar().value()
        posY = event.y()
        newY = s * (scrollY + posY) - posY
        self.scrollarea.verticalScrollBar().setValue(int(newY))

        self.blink(False)

    def mousePressEvent(self, event):
        assert isinstance(event, QMouseEvent)
        self.key_signal.emit(KEY_MOUSE, event.x(), event.y())
        self.blink(False)
        event.accept()

    def keyPressEvent(self, event):
        assert isinstance(event, QKeyEvent)
        self.key_signal.emit(int(event.key()), 0, 0)
        self.blink(False)
        if event.key() == Qt.Key_Escape:
            self.close()
            event.accept()

    def moveEvent(self, event):
        self.move_signal.emit()

    def contextMenuEvent(self, event):
        assert isinstance(event, QContextMenuEvent)

        self.blink(False)

        menu = QMenu()

        copy = menu.addAction("Copy to clipboard")

        reset = menu.addAction("Reset view")

        hq = menu.addAction("High quality")
        hq.setCheckable(True)
        if self.quality == Qt.SmoothTransformation:
            hq.setChecked(True)

        fixed = menu.addAction("Fixed size")
        fixed.setCheckable(True)
        if self.fixed_size:
            fixed.setChecked(True)

        rotate_right = menu.addAction("Rotate +")
        rotate_left = menu.addAction("Rotate -")

        save = menu.addAction("Save...")

        quit = menu.addAction("Close")

        action = menu.exec_(self.mapToGlobal(event.pos()))

        if action == quit:
            self.close()
        elif action == reset:
            self.setParams(1,0)
        elif action == hq:
            if self.quality == Qt.SmoothTransformation:
                self.quality = Qt.FastTransformation
            else:
                self.quality = Qt.SmoothTransformation
            self.setZoom(self.scale)
        elif action == rotate_right:
            rotation = (self.rotation + 90) % 360
            self.setRotation(rotation)
        elif action == rotate_left:
            rotation = (self.rotation + 270) % 360
            self.setRotation(rotation)
        elif action == save:
            filename, filter = QFileDialog.getSaveFileName(self, "Save image...", filter="*.png;;*.jpg;;*.bmp;;*.tiff;;*.gif", directory=self.last_save_dir)
            if filename:
                try:
                    if not str(filename).endswith(filter[1:]):
                        filename = filename + filter[1:]
                    PreviewWindow.last_save_dir = path.dirname(str(filename))
                    success = self.image.save(filename, quality=100)
                    if not success: raise Exception("unknown error")
                except Exception as e:
                    QMessageBox.critical(self, "Saving error", "Cannot save.\nError: {}".format(e.message))
                    print("Saving error:", e)
        elif action == fixed:
            if self.fixed_size:
                self.fixed_size = None
            else:
                self.fixed_size = self.size()
        elif action == copy:
            print("copy")
            clipboard = QApplication.instance().clipboard()
            clipboard.setPixmap(self.image)

    def blink(self, enable):
        if enable:
            self.blink_timer.start()
        else:
            self.blink_timer.stop()
            self.blink_widget.hide()

    @pyqtSlot()
    def blink_(self):
        if self.blink_widget.isHidden():
            self.blink_widget.show()
        else:
            self.blink_widget.hide()


class WindowManager(QObject):
    imshow_signal = pyqtSignal(dict)
    moveWindow_signal = pyqtSignal(str, int, int)
    positions_file = "~/.image_preview"

    def __init__(self):
        QObject.__init__(self)
        self.lock = Lock()
        self.windows = {}
        self.positions = {}
        self.positions_last_save = datetime.now()
        self.load_positions()
        self.key = KEY_NONE
        self.key_position = (0,0)
        self.key_lock = threading.Condition()
        self.imshow_signal.connect(self._imshow_slot)
        self.moveWindow_signal.connect(self._moveWindow)

    def imshow(self, winname, mat, **kwargs):
        args = {'winname': winname, 'mat': mat}
        args.update(kwargs)
        self.imshow_signal.emit(args)

    @pyqtSlot(dict)
    def _imshow_slot(self, kwargs):
        self._imshow(**kwargs)

    def _imshow(self, winname, mat, **kwargs):
        window = self.window(winname)
        assert isinstance(window, PreviewWindow)
        window.setImageAndParams(mat, **kwargs)

    def window(self, winname, **kwargs):
        winname = str(winname)
        with self.lock:
            if winname not in self.windows:
                position = kwargs.pop('position', 'auto')
                if position == 'auto':
                    position = self.positions.get(winname, self.find_best_place())
                window = self.windows[winname] = PreviewWindow(self, winname, position=position, **kwargs)
                window.key_signal.connect(self.key_slot)
                window.move_signal.connect(self.save_positions)
            return self.windows[winname]

    def waitKey(self, delay=0):
        while True:
            key, position = self.waitKeyMouse(delay)
            if key != KEY_MOUSE: return key

    def waitKeyMouse(self, delay=0):
        if delay <= 0: timeout = None
        else: timeout = delay * 0.001
        with self.key_lock:
            self.key = KEY_NONE
            self.key_lock.wait(timeout)
            # print("KEY", self.key, self.key_position)
            return self.key, self.key_position

    @pyqtSlot(int, int, int)
    def key_slot(self, key, x, y):
        with self.key_lock:
            self.key = key
            self.key_position = (x,y)
            self.key_lock.notify()

    def find_best_place(self):
        positions_x = [w.frameGeometry().x()+w.width() for w in self.windows.values() if w.isVisible()]
        if not positions_x: return None
        x = max(positions_x) + 4
        y = 50
        return x, y

    def moveWindow(self, winname, x, y):
        self.moveWindow_signal.emit(winname, x, y)

    @pyqtSlot(str, int, int)
    def _moveWindow(self, winname, x, y):
        self.window(winname).move(x, y)

    def load_positions(self):
        try:
            path = os.path.expanduser(self.positions_file)
            if os.path.isfile(path):
                with self.lock:
                    self.positions.update(json.load(open(path)))
        except Exception:
            print("WARN: Cannot load window positions")

    @pyqtSlot()
    def save_positions(self, force=False):
        if not force and datetime.now() - self.positions_last_save < timedelta(seconds=5):
            return
        path = os.path.expanduser(self.positions_file)
        with self.lock:
            for name, window in self.windows.items():
                assert isinstance(window, PreviewWindow)
                self.positions[name] = window.x(), window.y()
        with open(path, 'w') as f:
            json.dump(self.positions, f)
        self.positions_last_save = datetime.now()

    def __del__(self):
        self.save_positions(True)


class ManagerThread(Thread):
    def __init__(self):
        super(ManagerThread, self).__init__()
        self.ready = Event()
        self.terminate = False
        self.daemon = True
        self.manager = None
        self.app = None
        self.start()
        self.ready.wait()

    def run(self):
        self.manager = WindowManager()
        self.app = QApplication([])
        self.ready.set()
        while not self.terminate and self.is_main_thread_alive():
            self.app.exec_()
        self.app.closeAllWindows()

    def is_main_thread_alive(self):
        for i in threading.enumerate():
            if i.name == "MainThread":
                return i.is_alive()
        return None

    def imshow(self, winname, mat, **kwargs):
        return self.manager.imshow(winname, mat, **kwargs)

    def waitKey(self, delay=0):
        return self.manager.waitKey(delay)

    def waitKeyMouse(self, delay=0):
        return self.manager.waitKeyMouse(delay)

    def moveWindow(self, winname, x, y):
        return self.manager.moveWindow(winname, x, y)


class Manager:
    def __init__(self):
        self._manager = None
        self.lock = Lock()

    @property
    def manager(self):
        if not self._manager:
            self.create_manager()
        return self._manager

    def create_manager(self):
        with self.lock:
            if self._manager:
                return
            if QApplication.instance() is None:
                print("Creating threaded window manager")
                self._manager = ManagerThread()
            else:
                print("Creating normal window manager")
                self._manager = WindowManager()

    def imshow(self, winname, mat, **kwargs):
        return self.manager.imshow(winname, mat, **kwargs)

    def waitKey(self, time=0):
        return self.manager.waitKey(time)

    def waitKeyMouse(self, time=0):
        return self.manager.waitKeyMouse(time)

    def moveWindow(self, winname, x, y):
        return self.manager.moveWindow(winname, x, y)


manager = Manager()
imshow = manager.imshow
waitKey = manager.waitKey
moveWindow = manager.moveWindow
waitKeyMouse = manager.waitKeyMouse


def namedWindow(*a,**kw):
    pass


__ALL__ = ['imshow', 'waitKey', 'waitKeyMouse', 'moveWindow', 'namedWindow']



from .base import *


class GrabCut(NormalElement):
    name = "GrabCut"
    comment = "Foreground Extraction using GrabCut Algorithm"

    def get_attributes(self):
        return [Input("image"), Input("classes")], \
               [Output("classes")], \
               [ComboboxParameter('convert_input', [('None', ''), ('4-value', '4'), ('255-value', '255')], "Convert input", 1),
                ComboboxParameter('convert_output', [('None', ''), ('2-value', '2'), ('4-value', '4'), ('255 2-value', '255-2'), ('255 4-value', '255-4')], "Convert output", 1),
                IntParameter('iterations', "Iterations", 1, 1, 100)]

    def process_inputs(self, inputs, outputs, parameters):
        image = inputs['image'].value
        classes = inputs['classes'].value+0

        if parameters['convert_input']:
            if parameters['convert_input'] == '255':
                classes //= 51
            tmp = np.zeros_like(classes)
            tmp[classes == 0] = cv.GC_BGD
            tmp[classes == 1] = cv.GC_PR_BGD
            tmp[classes == 2] = cv.GC_PR_FGD
            tmp[classes >= 3] = cv.GC_FGD
            classes = tmp

        bgdmodel = np.zeros((1, 65), np.float64)
        fgdmodel = np.zeros((1, 65), np.float64)

        cv.grabCut(image, classes, None, bgdmodel, fgdmodel, parameters['iterations'], cv.GC_INIT_WITH_MASK)

        if parameters['convert_output'] == '2':
            tmp = np.zeros_like(classes)
            tmp[classes == cv.GC_BGD] = 0
            tmp[classes == cv.GC_PR_BGD] = 0
            tmp[classes == cv.GC_PR_FGD] = 1
            tmp[classes == cv.GC_FGD] = 1
            classes = tmp
        elif parameters['convert_output'] == '4':
            tmp = np.zeros_like(classes)
            tmp[classes == cv.GC_BGD] = 0
            tmp[classes == cv.GC_PR_BGD] = 1
            tmp[classes == cv.GC_PR_FGD] = 2
            tmp[classes == cv.GC_FGD] = 3
            classes = tmp
        elif parameters['convert_output'] == '255-2':
            tmp = np.zeros_like(classes)
            tmp[classes == cv.GC_BGD] = 0
            tmp[classes == cv.GC_PR_BGD] = 0
            tmp[classes == cv.GC_PR_FGD] = 255
            tmp[classes == cv.GC_FGD] = 255
            classes = tmp
        elif parameters['convert_output'] == '255-4':
            tmp = np.zeros_like(classes)
            tmp[classes == cv.GC_BGD] = 0
            tmp[classes == cv.GC_PR_BGD] = 85
            tmp[classes == cv.GC_PR_FGD] = 170
            tmp[classes == cv.GC_FGD] = 255
            classes = tmp

        outputs["classes"] = Data(classes)

register_elements_auto(__name__, locals(), "Segmentation", 6)



from .base import *


class ImageLoader(InputElement):
    name = "Image loader"
    comment = "Loads the image from disk"

    def get_attributes(self):
        return [], [Output("output")], [PathParameter("path", value="lena.jpg")]

    def process_inputs(self, inputs, outputs, parameters):
        path = parameters["path"]
        if path.endswith('lena.jpg'):
            path = CVLAB_DIR + "/images/lena.jpg"

        d = cv.imread(path)
        if d is not None:
            self.may_interrupt()
            outputs["output"] = Data(d)


class ImageSequenceLoader(InputElement):
    name = "Image Sequence loader"
    comment = "Loads a sequence of images from disk"

    def get_attributes(self):
        return [], [Output("output", "Sequence")], [MultiPathParameter("paths", value=[CVLAB_DIR+"/images/lena.jpg"])]

    def process(self):
        paths = self.parameters["paths"].get()
        images = len(paths)
        sequence = Sequence([Data() for _ in range(images)])
        self.outputs["output"].put(sequence)
        for path, data in zip(paths, sequence):
            self.may_interrupt()
            image = cv.imread(path)
            if image is not None:
                data.value = image


class ImageLoader3D(InputElement):
    name = "Image loader 3D"
    comment = "Loads multiple images as 3D image"

    def get_attributes(self):
        return [], [Output("output")], [MultiPathParameter("paths", value=[CVLAB_DIR+"/images/lena.jpg"]*10)]

    def process_inputs(self, inputs, outputs, parameters):
        paths = parameters["paths"]
        image = []

        for path in sorted(paths):
            slice = cv.imread(path)
            self.may_interrupt()
            image.append(slice)
            if slice.shape != image[0].shape:
                raise Exception("Inconsisten slice dimensions")

        image = np.array(image)
        outputs["output"] = Data(image)


class RecurrentSequenceLoader(InputElement):
    name = "Image Recurrent Sequence loader"
    comment = "Loads a recurrent sequence of images from disk (reflecting directory structure)"

    max_level = 5

    def __init__(self):
        super(RecurrentSequenceLoader, self).__init__()
        self.level = 0

    def get_attributes(self):
        return [], [Output("output", "Sequences")], [DirectoryParameter("directory", value="images")]

    def read_directory(self, directory):
        self.level += 1
        sequence = Sequence()
        for entry in os.listdir(directory):
            try:
                if not entry or entry[0] == '.': continue
                path = directory + "/" + entry
                if os.path.isdir(path):
                    if self.level >= self.max_level: continue
                    sequence.value.append(self.read_directory(path))
                else:
                    image = cv.imread(path)
                    if image is not None and image.shape:
                        data = ImageData(image)
                        sequence.value.append(data)
            except Exception:
                pass
        self.level -= 1
        return sequence

    def process(self):
        self.level = 0
        directory = self.parameters["directory"].get()
        data = self.read_directory(directory)
        self.outputs["output"].put(data)


class ImageSaver(NormalElement):
    name = "Image saver"
    comment = "Saves actual image (optionally makes a sequence from them)"

    def get_attributes(self):
        return [Input("input")], [], [SavePathParameter("path", value="")]

    def process_inputs(self, inputs, outputs, parameters):
        cv.imwrite(parameters["path"], inputs["input"].value)


class ArrayLoader(InputElement):
    name = "Array loader"
    comment = "Loads numpy array from disk"

    def get_attributes(self):
        return [], [Output("output")], [PathParameter("path", value=CVLAB_DIR+"/images/default.npy")]

    def process_inputs(self, inputs, outputs, parameters):
        d = np.load(parameters["path"])
        if d is not None:
            self.may_interrupt()
            outputs["output"] = Data(d)


class ArraySaver(NormalElement):
    name = "Array saver"
    comment = "Saves numpy array to disk"

    def get_attributes(self):
        return [Input("input")], [], [SavePathParameter("path", value="")]

    def process_inputs(self, inputs, outputs, parameters):
        np.save(parameters["path"], inputs["input"].value)


register_elements_auto(__name__, locals(), "Image IO", 2)



import os
import sys

from .version import __version__

CVLAB_DIR = os.path.abspath(__file__ + "/..").replace("\\", "/")


def init_libs():
    # system exception hook
    try:
        sys._excepthook = sys.excepthook
        def exception_hook(exctype, value, traceback):
            sys._excepthook(exctype, value, traceback)
            sys.exit(-1)
        sys.excepthook = exception_hook
    except Exception:
        pass

    # numpy error exceptions
    try:
        import numpy as np
        np.seterr(all='raise')
    except Exception:
        pass

    # PyQt exceptions workaround
    try:
        try:
            from PyQt5 import sip
        except ImportError:
            import sip
        sip.setdestroyonexit(False)
    except Exception:
        pass


def splash():
    from PyQt5.QtGui import QPixmap
    from PyQt5.QtWidgets import QSplashScreen, QLabel, QDesktopWidget
    from PyQt5.QtCore import QTimer

    pixmap = QPixmap(CVLAB_DIR + "/images/splash.png")
    splash = QSplashScreen(pixmap)

    splash.show()
    splash.raise_()

    QTimer.singleShot(5000, splash.close)

    return splash


def main(*args, **kwargs):
    init_libs()

    from PyQt5.QtWidgets import QApplication
    app = QApplication(sys.argv)

    s = splash()

    from .view.mainwindow import MainWindow
    main_window = MainWindow(app)
    ret_code = app.exec_()
    return ret_code


if __name__ == '__main__':
    exit(main())



from .base import *


class CodeElement(NormalElement):
    name = "Code element"
    comment = "Runs user-given Python code"

    def __init__(self):
        super(CodeElement, self).__init__()
        self.compiled_code_str = ""
        self.compiled_code = None
        self.memory = {}

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [TextParameter("code", value=u"import cv2 as cv\nimport numpy as np\n#your code here\nreturn None",
                              window_content="def fun(image=numpy.array, parameters={}, intpoint=func, memory={}):",
                              window_title="Code editor"),
                ComboboxParameter("split_channels", OrderedDict([("Channels", True),("Image", False)]), "What to process", 1)]

    def compile(self, code):
        code = str(code)
        self.compiled_code_str = code
        self.compiled_code = None
        code = code.replace("\r\n", "\n")
        code = u"def fun(image, parameters, intpoint, memory):\n\t" + code.replace("\n", "\n\t") + u"\n\treturn None"
        c = compile(code, "<string>", 'exec')
        loc = {}
        exec(c, loc)
        self.compiled_code = loc["fun"]

    def process_channels(self, inputs, outputs, parameters):
        outputs["output"] = Data(self.compiled_code(inputs["input"].value, parameters, self.may_interrupt, self.memory))

    def process_inputs(self, inputs, outputs, parameters):
        if self.compiled_code_str != parameters["code"]:
            self.compile(parameters["code"])
        if not self.compiled_code: return
        if parameters["split_channels"]:
            NormalElement.process_inputs(self, inputs, outputs, parameters)
        else:
            self.process_channels(inputs, outputs, parameters)

    def get_source(self):
        # fixme: this won't work with 'process channels' set
        name = self.__class__.__name__.lower() + str(abs(hash(self)))[-8:]
        fun_code = self.parameters["code"].get().replace("intpoint()", "").replace("\r\n","\n").replace("\n","\n\t").replace("\t","    ")
        source = """
# inner code for {name}
def {name}_fun(image, parameters, memory):
    {fun_code}

# memory for {name}
{name}_memory = {{}}

# general code for {name}
def {name}(inputs, outputs, parameters):
    image = inputs["input"].value
    parameters = {{}}
    global {name}_memory
    result = {name}_fun(image, parameters, {name}_memory)
    outputs["output"] = Data(result)
""".format(**locals())
        return name, source, []


class CodeElementEx(CodeElement):
    name = "Code element (extended version)"

    def get_attributes(self):
        return [Input("in1", optional=True), Input("in2", optional=True), Input("in3", optional=True),
                Input("in4", optional=True)], \
               [Output("o1"), Output("o2"), Output("o3"), Output("o4")], \
               [TextParameter("code",
                              value=u"import cv2 as cv\nimport numpy as np\n#your code here\nreturn None, None, None, "
                                    u"None",
                              window_content="def fun(in1, in2, in3, in4, parameters={}, intpoint=func, memory={}):",
                              window_title="Code editor"),
                ComboboxParameter("split_channels", OrderedDict([("Channels", True),("Image", False)]), "What to process", 1)]

    def compile(self, code):
        code = str(code)
        self.compiled_code_str = code
        self.compiled_code = None
        code = code.replace("\r\n", "\n")
        code = u"def fun(in1, in2, in3, in4, parameters, intpoint, memory):\n\t" + code.replace("\n",
                                                                                                "\n\t") + \
               u"\n\treturn None, None, None, None"
        c = compile(code, "<string>", 'exec')
        loc = {}
        exec(c, loc)
        self.compiled_code = loc["fun"]

    def process_channels(self, inputs, outputs, parameters):
        ins = [None] * 4
        for i in range(4):
            n = "in" + str(i + 1)
            if n in inputs and inputs[n]:
                ins[i] = inputs[n].value
        o = self.compiled_code(ins[0], ins[1], ins[2], ins[3], parameters, self.may_interrupt, self.memory)
        for i, v in enumerate(o):
            n = "o" + str(i + 1)
            outputs[n] = Data(v)

    def get_source(self):
        # fixme: this won't work with 'process channels' set
        name = self.__class__.__name__.lower() + "_" + str(abs(hash(self)))[-8:]
        fun_code = self.parameters["code"].get().replace("intpoint()", "").replace("\r\n","\n").replace("\n","\n\t").replace("\t","    ")
        source = """
# inner code for {name}
def {name}_fun(in1, in2, in3, in4, parameters, memory):
    {fun_code}

# memory for {name}
{name}_memory = {{}}

# general code for {name}
def {name}(inputs, outputs, parameters):
    ins = [None] * 4
    for i in range(4):
        n = "in" + str(i + 1)
        if n in inputs and inputs[n]:
            ins[i] = inputs[n].value
    o = {name}_fun(ins[0], ins[1], ins[2], ins[3], parameters, {name}_memory)
    for i, v in enumerate(o):
        n = "o" + str(i + 1)
        outputs[n] = Data(v)
""".format(**locals())
        return name, source, []


class CodeElementSequence(CodeElementEx, SequenceToSequenceElement):
    name = "Code element (sequence output)"
    num_outputs = 8

    def get_attributes(self):
        return [Input("inputs", multiple=True)], \
               [Output("output")], \
               [TextParameter("code",
                              value=u"import cv2 as cv\nimport numpy as np\n\n#your code here\n\nreturn []",
                              window_content="def fun(inputs, parameters={}, intpoint=func, memory={}):",
                              window_title="Code editor"),
                IntParameter("outputs", min_=1, max_=100, value=8)]

    def get_processing_units(self, inputs, parameters):
        self.num_outputs = parameters["outputs"]
        return SequenceToSequenceElement.get_processing_units(self, inputs, parameters)

    def process_inputs(self, inputs, outputs, parameters):
        if len(inputs["inputs"].value) == 0: raise Exception("Connect some inputs")

        if self.num_outputs != parameters['outputs']:
            self.recalculate(True, True, True, True)
            self.may_interrupt()

        if self.compiled_code_str != parameters["code"]:
            self.compile(parameters["code"])
        if not self.compiled_code: return

        inputs = inputs["inputs"].desequence_all()
        o = self.compiled_code(inputs, parameters, self.may_interrupt, self.memory)
        o = list(o)[:self.num_outputs]
        o += [None] * (self.num_outputs-len(o))
        outputs["output"] = Sequence([ImageData(d) for d in o])

    def compile(self, code):
        code = str(code)
        self.compiled_code_str = code
        self.compiled_code = None
        code = code.replace("\r\n", "\n")
        code = u"def fun(inputs, parameters, intpoint, memory):\n\t" + code.replace("\n", "\n\t") + \
               u"\n\treturn [None, None, None, None]"
        c = compile(code, "<string>", 'exec')
        loc = {}
        exec(c, loc)
        self.compiled_code = loc["fun"]


class InputCodeElement(NormalElement):
    name = "Input code element"
    comment = "Runs user-given Python code to provide input"
    example_code = \
        "# Your input generating code here, example:\n" + \
        "import cv2 as cv\n\n" + \
        "image = cv.imread('./images/lena.jpg')\n\n" + \
        "return image"

    def __init__(self):
        super(InputCodeElement, self).__init__()
        self.compiled_code_str = ""
        self.compiled_code = None
        self.memory = {}

    def get_attributes(self):
        return [], \
               [Output("output")], \
               [TextParameter("code", value=self.example_code,
                              window_content="def fun(parameters={}, intpoint=func, memory={}):",
                              window_title="Code editor")
                ]

    def compile(self, code):
        code = str(code)
        self.compiled_code_str = code
        self.compiled_code = None
        code = code.replace("\r\n", "\n")
        code = u"def fun(parameters, intpoint, memory):\n\t" + code.replace("\n", "\n\t") + u"\n\treturn None"
        c = compile(code, "<string>", 'exec')
        loc = {}
        exec(c, loc)
        self.compiled_code = loc["fun"]

    def process_inputs(self, inputs, outputs, parameters):
        if self.compiled_code_str != parameters["code"]:
            self.compile(parameters["code"])
        if self.compiled_code:
            outputs["output"] = Data(self.compiled_code(parameters, self.may_interrupt, self.memory))

    def get_source(self):
        name = self.__class__.__name__.lower() + str(abs(hash(self)))[-8:]
        fun_code = self.parameters["code"].get()\
            .replace("intpoint()", "")\
            .replace("\r\n", "\n")\
            .replace("\n", "\n\t")\
            .replace("\t", "    ")
        source = """
# inner code for {name}
def {name}_fun(parameters, memory):
    {fun_code}

# memory for {name}
{name}_memory = {{}}

# general code for {name}
def {name}(outputs, parameters):
    parameters = {{}}
    global {name}_memory
    result = {name}_fun(parameters, {name}_memory)
    outputs["output"] = Data(result)
""".format(**locals())
        return name, source, []


register_elements_auto(__name__, locals(), "Code", 10)



from distutils.util import strtobool

import cv2
from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSlot
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

from .. import CVLAB_DIR
from ..diagram.interface import *
from .mimedata import *
from . import image_preview
from . import config

ALLOW_UPSIZE = True


class StyledWidget(QWidget):
    def __init__(self):
        super().__init__()
        self.setAttribute(QtCore.Qt.WA_StyledBackground)
        self.setAttribute(QtCore.Qt.WA_StyleSheet)


class InOutConnector(StyledWidget):
    help = """\
Input / output connector anchor
Drag & drop to another element to connect their inputs/outputs"""

    def __init__(self, element, io_handle, is_input=False):
        super(InOutConnector, self).__init__()
        self.setObjectName("InOutButton")
        self.setAcceptDrops(True)
        self.setToolTip(self.help)
        self.is_input = is_input
        self.io_handle = io_handle
        self.element = element
        self.workarea = None
        # workaround: the hint should be a tooltip but it is not - because it can't be transparent for mouse (Qt bug)
        self.hint = QLabel(io_handle.name)
        self.hint.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents)
        self.hint.setVisible(False)
        self.has_connected_and_selected_wire = False
        element.element_relocated.connect(self.update_hint_position)

    def set_workarea(self, workarea):
        self.workarea = workarea
        self.hint.setParent(workarea)
        self.workarea.user_actions.cursor_line_started.connect(self.show_hint)
        self.workarea.user_actions.cursor_line_dropped.connect(self.hide_hint)

    def get_center_point(self):
        return self.pos() + self.element.pos() + QtCore.QPoint(self.width() // 2, self.height() // 2)

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.element.workarea.user_actions.cursor_line_started.emit()
            drag = QDrag(self)
            mime_data = QtCore.QMimeData()
            if self.is_input:
                mime_data.setText(Mime.OUTGOING_CONNECTION)
            else:
                mime_data.setText(Mime.INCOMING_CONNECTION)
            drag.setMimeData(mime_data)
            drag.setPixmap(QPixmap(1, 1))     # todo: can this hack be removed?
            drag.exec_()

    def dropEvent(self, e):
        source = e.source()
        if self.element is not source.element and self.is_input ^ source.is_input:
            diagram = self.element.workarea.diagram
            diagram.connect_io(self.io_handle, source.io_handle)
            e.accept()
        self.element.workarea.user_actions.cursor_line_dropped.emit()

    def dragEnterEvent(self, e):
        mime = e.mimeData().text()
        if ((mime == Mime.INCOMING_CONNECTION and self.is_input) or
                (mime == Mime.OUTGOING_CONNECTION and not self.is_input)):
            e.accept()

    def dragMoveEvent(self, e):
        mime = e.mimeData().text()
        if mime in [Mime.INCOMING_CONNECTION, Mime.OUTGOING_CONNECTION]:
            if e.source().element is not self.element:
                points = [e.source().get_center_point(), self.get_center_point()]
            else:
                points = [e.source().get_center_point(), e.pos() + self.element.pos()]
            if not self.is_input:
                points.reverse()
            self.element.workarea.user_actions.cursor_line_moved.emit(tuple(points))

    def paintEvent(self, e):
        super(InOutConnector, self).paintEvent(e)
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing, True)
        if not self.is_input and len(self.io_handle.connected_to):
            pen, point = self.prepare_drawing_details()
            self.workarea.wire_tools.draw_start_symbol(painter, point, pen)
        if self.is_input and len(self.io_handle.connected_from):
            pen, point = self.prepare_drawing_details()
            self.workarea.wire_tools.draw_end_symbol(painter, point, pen)

    def prepare_drawing_details(self):
        wire_tools = self.workarea.wire_tools
        pen = wire_tools.pen_selected if self.has_connected_and_selected_wire else wire_tools.pen_regular
        point = QtCore.QPoint(self.width() // 2, self.width() // 2)
        return pen, point

    @pyqtSlot()
    def show_hint(self):
        self.hint.setVisible(True) #to musi byc w tej kolejnosci, bo inaczej hint.width jest losowe :/
        self.update_hint_position()
        self.hint.raise_()

    @pyqtSlot()
    def hide_hint(self):
        self.hint.setVisible(False)

    @pyqtSlot()
    def update_hint_position(self):
        if self.is_input:
            delta = -self.hint.width() - 5
        else:
            delta = self.width() + 5
        self.hint.move(self.pos() + self.parent().pos() + QtCore.QPoint(delta, 0))


class PreviewsContainer(StyledWidget):
    help = """\
Element outputs preview
Mouse wheel - resize the previews
Double click - open image preview in separate window or expand/collapse text preview"""

    def __init__(self, element, outputs):
        super(PreviewsContainer, self).__init__()
        self.element = element
        self.outputs = outputs
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.preview_size = 120
        self.setObjectName("OutputsPreview")
        self.setVisible(False)
        self.previews = []
        self.create_previews(layout)
        self.setLayout(layout)
        self.element.state_changed.connect(self.update)
        self.image_dialogs_count = 0
        self.setToolTip(self.help)

    def wheelEvent(self, event):
        assert isinstance(event, QWheelEvent)
        if event.modifiers() != QtCore.Qt.NoModifier:
            event.ignore()
            return
        new_size = self.preview_size
        up = event.angleDelta().y() > 0
        for _ in range(abs(event.angleDelta().y() // 80)):
            if up:
                new_size *= 1.125
            else:
                new_size /= 1.125
        self.resize_previews(new_size)
        event.accept()

    def resize_previews(self, new_size):
        out_size = np.clip(int(new_size), 64, 2048)
        self.preview_size = int(out_size)
        for preview in self.previews:
            preview.preview_size = out_size
        self.update()

    def create_previews(self, layout):
        for output in self.outputs:
            if not output.preview_enabled:
                continue
            preview = OutputPreview(output, self)
            self.previews.append(preview)
            layout.addLayout(preview)

    @pyqtSlot()
    def update(self):
        if not hasattr(self, "element"):
            # fixme: tymczasowy hack, bo leci w tym miejscu wyjątek, nie wiem czemu!
            print("Error: ", self, " nie posiada atrybutu 'element'!")
            return
        self.element.state_notified = False  # todo: this must be called by all slots connected to self.element.state_changed
        state = self.element.state
        #if state == self.element.STATE_READY:
        #    self.update_previews(state)
        #else:
        #    self.set_outdated()
        self.update_previews(state) #todo: czy tak, czy lepiej powyzsze z komentarza?

    def switch_visibility(self, value):
        if value is None:
            value = not self.isVisible()
        if not self.isVisible():
            self.force_update()
        self.setVisible(value)

    def set_outdated(self):
        for preview in self.previews:
            preview.set_outdated()

    def update_previews(self, state):
        if self.isVisible() or self.image_dialogs_count:
            for preview in self.previews:
                preview.update()

    def force_update(self):
        for preview in self.previews:
            preview.update(True)


class ActionImage(QLabel):

    DATA_TYPE_IMAGE = 0
    DATA_TYPE_TEXT = 1
    DATA_TYPE_VALUE = 3

    def __init__(self, image_preview):
        super(ActionImage, self).__init__()
        self.image_preview = image_preview
        self.previews_container = image_preview.previews_container
        self.id = len(image_preview.previews)
        self.element = self.previews_container.element
        #todo: better Element number than object_id would be appreciated here
        self.name = "{} {}, Output {}, Image {}".format(self.element.name, str(self.element.object_id), image_preview.output.name, str(self.id))
        self.__connected = False
        self.image_dialog = None
        self.data_type = ActionImage.DATA_TYPE_IMAGE
        self.number_output_helper = NumberOutputHelper()
        self.setMargin(0)
        self.prepare_actions()
        self.setObjectName("OutputPreview")
        self.text_preview_expanded = False

    def set_image(self, arr):
        # remember not to modify arr !!!
        if self.data_type != ActionImage.DATA_TYPE_IMAGE:
            self.prepare_actions()
        self.data_type = ActionImage.DATA_TYPE_IMAGE
        if isinstance(arr, np.ndarray):
            arr = self.preprocess_array(arr)
        if arr is None:
            pass
            # self.setPixmap(self.image_preview.default_image)  # todo: na pewno to chcemy? moze to nam opozniac interfejs!
        elif isinstance(arr, np.ndarray):
            qpix = image_preview.array_to_pixmap(arr)
            qpix_scaled = self.scale_pixmap(qpix)
            self.setPixmap(qpix_scaled)
            if self.image_dialog is not None:
                image_preview.imshow(self.name, qpix, show=False)

    def set_text(self, arr):
        if self.data_type != ActionImage.DATA_TYPE_TEXT:
            self.prepare_actions(enable=False)
        self.data_type = ActionImage.DATA_TYPE_TEXT
        self.setText(arr)
        self.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)

    def set_bool(self, value):
        if self.data_type != ActionImage.DATA_TYPE_VALUE:
            self.prepare_actions(enable=False)
        self.data_type = ActionImage.DATA_TYPE_VALUE
        qpix = self.number_output_helper.get_output(value)
        qpix = self.scale_pixmap(qpix)
        self.setPixmap(qpix)

    def preprocess_array(self, arr):
        if arr.dtype == np.uint16:
            arr = cv.convertScaleAbs(arr, alpha=255.0/65535.0)
        elif arr.dtype != np.uint8:
            if arr.dtype in [np.float, np.float32, np.float64]:
                arr = arr * 255.
            if arr.min() < 0:
                arr = arr // 2 + 127
            arr = arr.clip(0, 255)
            arr = np.uint8(arr)
        if len(arr.shape) == 2:
            return cv.cvtColor(arr, cv.COLOR_GRAY2BGRA)
        elif len(arr.shape) == 3 and arr.shape[2] == 3:
            return cv.cvtColor(arr, cv.COLOR_BGR2BGRA)
        else:
            return None

    def scale_pixmap(self, qpix):
        hq = bool(strtobool(config.ConfigWrapper.get_settings().get_with_default(config.VIEW_SECTION,
                                                                                 config.VIEW_HQ_OPTION)))
        quality = QtCore.Qt.SmoothTransformation if hq else QtCore.Qt.FastTransformation
        size = self.previews_container.preview_size
        if not ALLOW_UPSIZE and size > max(qpix.width(), qpix.height()):
            size = max(qpix.width(), qpix.height())
        if qpix.width() > qpix.height():
            result = qpix.scaledToWidth(size, quality)
        else:
            result = qpix.scaledToHeight(size, quality)
        return result

    def mouseDoubleClickEvent(self, mouse_event):
        if self.data_type == self.DATA_TYPE_IMAGE:
            self.open_image_dialog()
        if self.data_type == self.DATA_TYPE_TEXT:
            self.text_preview_expanded = not self.text_preview_expanded
            self.image_preview.update()

    @pyqtSlot()
    def open_image_dialog(self):
        if not self.__connected and self.element.diagram is not None:
            self.element.diagram.element_deleted.connect(self.on_element_destroy)
            self.__connected = True
        if self.image_dialog is None:
            image = self.image_preview.get_preview_objects()[self.id]
            self.image_dialog = image_preview.manager.manager.window(self.name, image=image, position='cursor')
            self.image_dialog.setImage(image)
            settings = config.ConfigWrapper.get_settings()
            if bool(strtobool(settings.get_with_default(config.VIEW_SECTION, config.PREVIEW_ON_TOP_OPTION))):
                flags = self.image_dialog.windowFlags()
                flags |= QtCore.Qt.WindowStaysOnTopHint
                self.image_dialog.setWindowFlags(flags)
                self.image_dialog.showNormal()
            self.image_dialog.installEventFilter(self)
            self.previews_container.image_dialogs_count += 1

    def prepare_actions(self, enable=True):
        if enable:
            self.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)
            action = QAction('Open in window', self)
            action.triggered.connect(self.open_image_dialog)
            self.addAction(action)
        else:
            self.removeAction(self.actions()[0])

    def close_image_dialog(self):
        if self.image_dialog is not None:
            self.image_dialog.close()
            self.image_dialog = None
            self.previews_container.image_dialogs_count -= 1
            if self.previews_container.image_dialogs_count < 0:
                self.previews_container.image_dialogs_count = 0
                # todo: report error here: reaching here would mean unexpected circumstances in counting spawning and closing dialogs

    def eventFilter(self, source, event):
        if event.type() == QtCore.QEvent.Close and self.image_dialog is not None:
            self.close_image_dialog()
            return True
        return False

    @pyqtSlot(Element)
    def on_element_destroy(self, element):
        if element is self.element and self.image_dialog is not None:
            self.close_image_dialog()

    def deleteLater(self):
        QObject.deleteLater(self)
        if self.image_dialog is not None:
            self.close_image_dialog()


class OutputPreview(QHBoxLayout):
    default_image = None
    preview_callbacks = [
        (np.ndarray, ActionImage.set_image),
        (str, ActionImage.set_text),
        (bool, ActionImage.set_bool)
    ]

    def __init__(self, output, previews_container):
        super(OutputPreview, self).__init__()
        if not self.default_image:
            self.default_image = QPixmap(CVLAB_DIR + "/images/default.png")
        self.output = output
        self.previews_container = previews_container
        self.setAlignment(QtCore.Qt.AlignCenter)
        self.setContentsMargins(0, 0, 0, 0)
        self.base_spacing = 4
        self.previews = []
        self.previews.append(ActionImage(self))
        self.img = self.default_image
        self.previews[0].setPixmap(self.img)
        self.addWidget(self.previews[0])

    def update(self, forced=False):
        objects = self.get_preview_objects()
        if not objects:
            objects = [None]
        self.adjust_number_of_previews(objects)
        for i, obj in enumerate(objects):
            if forced or self.previews_container.isVisible() or self.previews[i].image_dialog is not None:
                for _type, callback in self.preview_callbacks:
                    if isinstance(obj, _type):
                        try:
                            callback(self.previews[i], obj)
                        except Exception as e:
                            print("Cannot update preview: {}".format(e))

    def get_preview_objects(self):
        return self.output.get().desequence_all()

    def adjust_number_of_previews(self, preview_images):
        while len(preview_images) > len(self.previews):
            new_label = ActionImage(self)
            self.addWidget(new_label)
            self.previews.append(new_label)
        while len(preview_images) < len(self.previews):
            label_out = self.previews[-1]
            self.previews.pop(-1)
            # todo: again - a memory leak?
            label_out.deleteLater()

    def set_outdated(self):
        # todo: implement this
        pass


class NumberOutputHelper:

    def __init__(self):
        true_ = np.ones((11, 11, 4), dtype=np.uint8) * 255
        cv2.line(true_, (5, 3), (5, 7), color=(0, 0, 0, 255))
        cv2.line(true_, (4, 4), (4, 4), color=(0, 0, 0, 255))
        self.true_ = image_preview.array_to_pixmap(true_)
        false_ = np.zeros((11, 11, 4), dtype=np.uint8) + np.array([0, 0, 0, 255], dtype=np.uint8)
        cv2.rectangle(false_, (4, 3), (6, 7), color=(255, 255, 255, 255), thickness=1)
        self.false_ = image_preview.array_to_pixmap(false_)

    def get_output(self, value):
        if isinstance(value, bool):
            return self.true_ if value else self.false_


class ElementStatusBar(StyledWidget):
    def __init__(self, element):
        super(ElementStatusBar, self).__init__()
        self.element = element
        # self.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
        self.default_message = "Element unset."
        self.message = QLabel(self.default_message)
        self.message.setWordWrap(True)
        self.message.setObjectName("ElementStatusLabel")
        # self.message.setSizePolicy(QSizePolicy.Ignored,QSizePolicy.Ignored)
        self.timings = QLabel("")
        self.timings.setVisible(False)
        self.timings.setObjectName("ElementStatusLabel")
        self.timings.setAlignment(QtCore.Qt.AlignRight)
        # self.timings.setSizePolicy(QSizePolicy.Ignored,QSizePolicy.Ignored)
        hb = QHBoxLayout()
        hb.setContentsMargins(0, 0, 0, 0)
        hb.setSpacing(0)
        hb.addWidget(self.message)
        hb.addWidget(self.timings)
        self.setLayout(hb)
        self.element.state_changed.connect(self.update)

    def set_status(self, message=""):
        # Todo: the below is slow enough to slightly freeze the GUI; QT SIGNALS (state_changed) might be the cause
        if message != "":
            self.message.setText(message)
        else:
            self.message.setText(self.default_message)
            # Todo: self.message.adjustSize() might be required here
        self.display_processing_time()

    def display_processing_time(self):
        if not hasattr(self.element, "processing_time_info"): return  # fixme: tymczasowy hack...
        time_info = self.element.processing_time_info
        if time_info is None: return
        if self.element.state == self.element.STATE_READY and time_info is not None:
            text = self.prepare_time_label()
            tooltip = self.prepare_tooltip()
            self.timings.setText(text)
            self.timings.setToolTip(tooltip)
            self.timings.setVisible(True)
        else:
            self.timings.setVisible(False)

    def prepare_time_label(self):
        time_info = self.element.processing_time_info
        unit_element_time = self.get_text_for_milis(time_info.work_time_per_unit)
        unit_total_time = self.get_text_for_milis(time_info.total_work_time_per_unit)
        return unit_element_time + " / " + unit_total_time

    def prepare_tooltip(self):
        time_info = self.element.processing_time_info
        element_time = self.get_text_for_milis(time_info.work_time)
        total_time = self.get_text_for_milis(time_info.total_work_time)
        if time_info.work_time == time_info.work_time_per_unit:
            tooltip = "%s - Element processing time\n%s - Total processing time" % (element_time, total_time)
            return tooltip
        else:
            unit_element_time = self.get_text_for_milis(time_info.work_time_per_unit)
            unit_total_time = self.get_text_for_milis(time_info.total_work_time_per_unit)
            tooltip = "%s - Unit element processing time\n%s - Unit total processing time\n%s - Element processing time\n%s - Total processing time" \
                      % (unit_element_time, unit_total_time, element_time, total_time)
            return tooltip

    def get_text_for_milis(self, milis):
        return (str(milis) + " ms") if milis < 1000 else ("%.2f" % (milis / 1000.0) + " s")

    @pyqtSlot()
    def update(self):
        if not hasattr(self, "element"): return  # fixme: tymczasowy hack, bo leci w tym miejscu wyjątek, nie wiem czemu!
        self.element.state_notified = False  # todo: this must be called by all slots connected to self.element.state_changed
        self.set_status(self.element.message)



from matplotlib.figure import Figure
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg

from .base import *


class ImagePreview(NormalElement):
    name = "Image preview"
    comment = "Shows an image using selected presentation method"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [ComboboxParameter("type", [
                ("No change", 0),
                ("Truncate to 0-255", 10),
                ("Truncate to 0-1", 20),
                ("Scale contrast to 0-1", 21),
                ("Mean -> 0.5 max/min -> 1/0", 22),
                ("Divide by 255.0", 23),
                ("0 -> 0.5, max/min -> 1/0", 30),
               ])]

    def process_inputs(self, inputs, outputs, parameters):
        i = inputs["input"].value
        t = parameters["type"]
        o = None
        if t == 0:
            o = i
        elif t == 10:
            o = i.clip(0, 255).astype(np.uint8)
        elif t == 20:
            o = i.astype(np.float32).clip(0.0, 1.0)
        elif t == 21:
            o = i.astype(np.float32)
            min_, max_, _, _ = cv.minMaxLoc(o.flatten())
            if min_ == max_:
                o = np.zeros(o.shape) + 0.5
            else:
                o = (o-min_)/(max_-min_)+min_
        elif t == 22:
            o = i.astype(np.float32)
            min_, max_, _, _ = cv.minMaxLoc(o.flatten())
            if min_ == max_:
                o = np.zeros(o.shape) + 0.5
            else:
                if len(o.shape) > 2:
                    mean = cv.mean(cv.mean(o)[:3])[0]
                else:
                    mean = cv.mean(o)[0]
                scale = 0.5/max(max_-mean, mean-min_)
                o = (o-mean)*scale+0.5
        elif t == 23:
            o = i / 255.
        elif t == 30:
            o = i.astype(np.float32)
            min_, max_, _, _ = cv.minMaxLoc(o.flatten())
            if min_ == max_:
                o = np.zeros(o.shape) + 0.5
            else:
                scale = 0.5/max(max_,-min_)
                o = o*scale+0.5
        outputs["output"] = Data(o)


class ImagePreview3D(ImagePreview):
    name = "Image preview 3D"
    comment = "Shows selected slice from 3D image"

    def get_attributes(self):
        attributes = super(ImagePreview3D, self).get_attributes()
        attributes[2].append(ComboboxParameter("axis", [("Z",0),("Y",1),("X",2)]))
        attributes[2].append(FloatParameter("slice", min_=0, max_=1, step=0.01))
        return attributes
    
    def process_inputs(self, inputs, outputs, parameters):
        image = inputs["input"].value
        axis = parameters["axis"]
        slice = int(round(parameters["slice"] * (image.shape[axis]-1)))

        if axis == 0: slice = image[slice, ...]
        elif axis == 1: slice = image[:, slice, ...]
        elif axis == 2: slice = image[:, :, slice, ...]
        else: raise Exception("Unknown 'axis' parameter")

        inputs["input"].value = slice
        super(ImagePreview3D, self).process_inputs(inputs, outputs, parameters)


class Plot3d(NormalElement):
    name = "Plot 3D (wireframe)"
    comment = "Simple plot of 3D wireframe"

    def __init__(self):
        super(Plot3d, self).__init__()
        self.figure = Figure(figsize=(4, 4), dpi=90, facecolor=(1, 1, 1), edgecolor=(0, 0, 0))
        self.axes = self.figure.add_subplot(111, projection='3d')
        # self.axes.hold(False)
        self.plot_widget = FigureCanvasQTAgg(self.figure)
        self.axes.mouse_init()
        self.layout().addWidget(self.plot_widget)

    def get_attributes(self):
        return [Input("input")], [], []

    def process_inputs(self, inputs, outputs, parameters):
        i = inputs["input"].value
        X = i[0]
        Y = i[1]
        Z = i[2]

        self.axes.cla()
        self.axes.plot_wireframe(X, Y, Z)
        self.plot_widget.draw()
        self.plot_widget.updateGeometry()


register_elements_auto(__name__, locals(), "Visualization", 3)



from PyQt5.QtCore import pyqtSignal, pyqtSlot, Qt
from PyQt5.QtGui import QCursor
from PyQt5.QtWidgets import QLineEdit, QDoubleSpinBox, QSpinBox, QApplication


class NumberLineEdit(QLineEdit):
    def __init__(self):
        super().__init__()
        self.pressed_pos = None

    def mousePressEvent(self, event):
        QApplication.instance().setOverrideCursor(Qt.BlankCursor)
        self.pressed_pos = event.globalPos()
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        QApplication.instance().restoreOverrideCursor()
        self.pressed_pos = None
        super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event):
        if not self.pressed_pos: return
        new_pos = event.globalPos()
        QCursor.setPos(self.pressed_pos)
        difference = new_pos - self.pressed_pos
        x, y = difference.x(), difference.y()
        if abs(x) >= abs(y) or abs(x)+abs(y)<=2:
            add = difference.x() / 10000.0
        else:
            add = -difference.y() / 500.0
        self.parent().mouseSpin.emit(add, 1.0)


class DoubleSpinBoxEx(QDoubleSpinBox):
    mouseSpin = pyqtSignal(float, float)

    def __init__(self):
        super().__init__()
        self.pressed_pos = None
        self.setLineEdit(NumberLineEdit())
        self.mouseSpin.connect(self.change_value)

    @pyqtSlot(float, float)
    def change_value(self, add, mul):
        value = self.value()
        if value < 0: mul = 1/mul
        value = value * mul + add * (self.maximum()-self.minimum())
        self.setValue(value)


class SpinBoxEx(QSpinBox):
    mouseSpin = pyqtSignal(float, float)

    def __init__(self):
        super().__init__()
        self.pressed_pos = None
        self.setLineEdit(NumberLineEdit())
        self.mouseSpin.connect(self.change_value)
        self.valueChanged.connect(self.value_changed_slot)
        self.float_value = None

    @pyqtSlot(int)
    def value_changed_slot(self, value):
        self.float_value = value

    @pyqtSlot(float, float)
    def change_value(self, add, mul):
        value = self.float_value or 0.0
        if value < 0: mul = 1/mul
        value = value * mul + add * (self.maximum()-self.minimum())
        self.setValue(int(round(value)))
        self.float_value = value



#
# Creating your own Element for CV Lab
#

# import whatever libraries you want

import cv2 as cv

# also import this
from cvlab.diagram.elements.base import *


# create a class which inherits NormalElement
class MySampleElement(NormalElement):
    name = "My sample element"
    comment = "My first self-created element, hurray! :)"
    package = "My private elements"

    def get_attributes(self):
        return [Input("input", name="Input")], \
               [Output("output-1", name="First output"), Output("output-2", name="Second output")], \
               [IntParameter("param", "Some integer parameter", min_=1, max_=255)]

    def process_inputs(self, inputs, outputs, parameters):
        # read the parameters and inputs
        image = inputs["input"].value
        param = parameters["param"]

        o1 = cv.blur(image, (param, param))
        self.may_interrupt()  # allow interruptions here
        o2 = image + param

        # set the outputs
        outputs["output-1"] = Data(o1)
        outputs["output-2"] = Data(o2)


# register all elements defined above in the palette
register_elements_auto(__name__, locals(), "My private elements", 1)




 # ############################################################################
 #
 # Copyright (c) Microsoft Corporation. 
 #
 # Available under the Microsoft PyKinect 1.0 Alpha license.  See LICENSE.txt
 # for more information.
 #
 # ###########################################################################/

"""Contains low-level implementation details which aren't consumed directly by users of the Kinect API"""

import ctypes
from cvlab_experimental.thirdparty.pykinect.nui import KinectError, _NUIDLL
from cvlab_experimental.thirdparty.pykinect.nui.structs import (ImageFrame, ImageResolution, ImageType,
                                  ImageViewArea,  SkeletonFrame, 
                                  TransformSmoothParameters, _Enumeration)

_SEVERITY_ERROR = 1

class _KinectHRESULT(ctypes._SimpleCData):
    """performs error checking and returns a custom error message for kinect HRESULTs"""
    _type_ = "l"
    @staticmethod
    def _check_retval_(error):
        if error < 0:
            err = KinectError(error, 22)
            error = error + 0xffffffff
            error_msg = _KINECT_ERRORS.get(error, '')
            if error_msg is not None:
                err.strerror = err.message = error_msg
        
            raise err

def _HRESULT_FROM_WIN32(error):
    return 0x80070000 | error

def _MAKE_HRESULT(sev, fac, code):
    return (sev << 31) | (fac << 16) | code


class _PropsIndex(_Enumeration):
    INDEX_UNIQUE_DEVICE_NAME = 0
    INDEX_LAST  = 1                     # don't use!


class _PropType(_Enumeration):
    UNKNOWN = 0   # don't use
    UINT = 1      # no need to return anything smaller than an int
    FLOAT = 2
    BSTR = 3      # returns new BSTR. Use SysFreeString( BSTR ) when you're done
    BLOB = 4


##############################################################################
##
## Define NUI error codes derived from win32 errors
##

_E_NUI_DEVICE_NOT_CONNECTED  = _HRESULT_FROM_WIN32(1167)
_E_NUI_DEVICE_NOT_READY      = _HRESULT_FROM_WIN32(21)
_E_NUI_ALREADY_INITIALIZED   = _HRESULT_FROM_WIN32(1247)
_E_NUI_NO_MORE_ITEMS         = _HRESULT_FROM_WIN32(259)

##
## Define NUI specific error codes
##

_FACILITY_NUI = 0x301
_E_NUI_FRAME_NO_DATA                     = _MAKE_HRESULT(_SEVERITY_ERROR, _FACILITY_NUI, 1)
_E_NUI_STREAM_NOT_ENABLED                = _MAKE_HRESULT(_SEVERITY_ERROR, _FACILITY_NUI, 2)
_E_NUI_IMAGE_STREAM_IN_USE               = _MAKE_HRESULT(_SEVERITY_ERROR, _FACILITY_NUI, 3)
_E_NUI_FRAME_LIMIT_EXCEEDED              = _MAKE_HRESULT(_SEVERITY_ERROR, _FACILITY_NUI, 4)
_E_NUI_FEATURE_NOT_INITIALIZED           = _MAKE_HRESULT(_SEVERITY_ERROR, _FACILITY_NUI, 5)
_E_NUI_DATABASE_NOT_FOUND                = _MAKE_HRESULT(_SEVERITY_ERROR, _FACILITY_NUI, 13)
_E_NUI_DATABASE_VERSION_MISMATCH         = _MAKE_HRESULT(_SEVERITY_ERROR, _FACILITY_NUI, 14)

_KINECT_ERRORS = {
    _E_NUI_DEVICE_NOT_CONNECTED : 'Device not connected',
    _E_NUI_DEVICE_NOT_READY : 'Device not ready',
    _E_NUI_ALREADY_INITIALIZED : 'Device already initialized',
    _E_NUI_NO_MORE_ITEMS : 'No more items',
    _E_NUI_FRAME_NO_DATA : 'Frame has no data',
    _E_NUI_STREAM_NOT_ENABLED : 'Stream is not enabled',
    _E_NUI_IMAGE_STREAM_IN_USE : 'Image stream is already in use',
    _E_NUI_FRAME_LIMIT_EXCEEDED : 'Frame limit exceeded',
    _E_NUI_FEATURE_NOT_INITIALIZED : 'Feature not initialized',
    _E_NUI_DATABASE_NOT_FOUND : 'Database not found',
    _E_NUI_DATABASE_VERSION_MISMATCH : 'Database version mismatch',
}

try:
    c_bool = ctypes.c_bool
except:
    c_bool = ctypes.c_uint

_kernel32 = ctypes.WinDLL('kernel32')
_CreateEvent = _kernel32.CreateEventW
_CreateEvent.argtypes = [ctypes.c_voidp, ctypes.c_uint, ctypes.c_bool, ctypes.c_wchar_p]
_CreateEvent.restype = ctypes.c_voidp

_CloseHandle = _kernel32.CloseHandle
_CloseHandle.argtypes = [ctypes.c_voidp]
_CloseHandle.restype = c_bool

_WaitForSingleObject = _kernel32.WaitForSingleObject 
_WaitForSingleObject.argtypes = [ctypes.c_voidp, ctypes.c_uint32]
_WaitForSingleObject.restype = ctypes.c_uint32

_WaitForMultipleObjects = _kernel32.WaitForMultipleObjects 
_WaitForMultipleObjects.argtypes = [ctypes.c_uint32, ctypes.POINTER(ctypes.c_voidp), ctypes.c_uint, ctypes.c_uint32]
_WaitForMultipleObjects.restype = ctypes.c_uint32

_WAIT_OBJECT_0 = 0
_INFINITE = 0xffffffff

_oleaut32 = ctypes.WinDLL('oleaut32')
_SysFreeString = _oleaut32.SysFreeString
_SysFreeString.argtypes = [ctypes.c_voidp]
_SysFreeString.restype = ctypes.HRESULT

class _NuiInstance(ctypes.c_voidp):
    """this interface duplicates exactly the public DLL NUI**** methods that
       work on just device #0. If you want to work with multiple devices,
       use these methods off the INuiInstance, after getting a INuiInstance * from
       the multiple-device methods below"""

    # vtable
    _NuiInitialize = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.c_uint32)(3, 'NuiInitialize')
    _NuiShutdown = ctypes.WINFUNCTYPE(None)(4, 'NuiShutdown')
    _NuiSetFrameAndEvent = ctypes.WINFUNCTYPE(_KinectHRESULT, ctypes.c_voidp, ctypes.c_uint32)(5, 'NuiSetFrameEndEvent')
    _NuiImageStreamOpen = ctypes.WINFUNCTYPE(_KinectHRESULT, ImageType, ImageResolution, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_voidp, ctypes.POINTER(ctypes.c_voidp))(6, 'NuiImageStreamOpen')
    _NuiImageStreamSetImageFrameFlags = 7
    _NuiImageStreamGetImageFrameFlags = 8
    _NuiImageStreamGetNextFrame = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.c_voidp, ctypes.c_uint32, ctypes.POINTER(ImageFrame))(9, 'NuiImageStreamGetNextFrame')
    _NuiImageStreamReleaseFrame = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.c_voidp, ctypes.POINTER(ImageFrame))(10, 'NuiImageStreamReleaseFrame')
    _NuiImageGetColorPixelCoordinatesFromDepthPixel = ctypes.WINFUNCTYPE(ctypes.HRESULT, 
        ImageResolution, 
        ctypes.POINTER(ImageViewArea), 
        ctypes.c_long, 
        ctypes.c_long, 
        ctypes.c_uint16, 
        ctypes.POINTER(ctypes.c_long), 
        ctypes.POINTER(ctypes.c_long))(11, 'NuiImageGetColorPixelCoordinatesFromDepthPixel')
    _NuiImageGetColorPixelCoordinatesFromDepthPixelAtResolution = 12
    _NuiImageGetColorPixelCoordinateFrameFromDepthPixelFrameAtResolution = 13
    _NuiCameraElevationSetAngle = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.c_long)(14, 'NuiCameraElevationSetAngle')
    _NuiCameraElevationGetAngle = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.POINTER(ctypes.c_long))(15, 'NuiCameraElevationGetAngle')
    _NuiSkeletonTrackingEnable = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.c_voidp, ctypes.c_uint32)(16, 'NuiSkeletonTrackingEnable')
    _NuiSkeletonTrackingDisable = ctypes.WINFUNCTYPE(ctypes.HRESULT)(17, 'NuiSkeletonTrackingDisable')
    _NuiSkeletonSetTrackedSkeletons = 18
    _NuiSkeletonGetNextFrame = ctypes.WINFUNCTYPE(_KinectHRESULT, ctypes.c_uint32, ctypes.POINTER(SkeletonFrame))(19, 'NuiSkeletonGetNextFrame')
    _NuiTransformSmooth = ctypes.WINFUNCTYPE(ctypes.HRESULT, ctypes.POINTER(SkeletonFrame), ctypes.POINTER(TransformSmoothParameters))(20, 'NuiTransformSmooth')
    _NuiGetAudioSource = 21
    _InstanceIndex = ctypes.WINFUNCTYPE(ctypes.c_int)(22, 'InstanceIndex')
    _NuiDeviceConnectionId = 23
    _NuiUniqueId = 24
    _NuiAudioArrayId = 25
    _NuiStatus = 26
    _NuiInitializationFlags = 27

    def InstanceIndex(self):
        """which instance # was it created with, in MSR_NuiCreateInstanceByIndex( )/etc?"""
        return _NuiInstance._InstanceIndex(self)

    def NuiInitialize(self, dwFlags = 0):
        _NuiInstance._NuiInitialize(self, dwFlags)

    def NuiShutdown(self):
        return _NuiInstance._NuiShutdown(self)
        
    def NuiImageStreamOpen(self, eImageType, eResolution, dwImageFrameFlags_NotUsed, dwFrameLimit, hNextFrameEvent = 0):
        res = ctypes.c_voidp()
        _NuiInstance._NuiImageStreamOpen(self, eImageType, eResolution, dwImageFrameFlags_NotUsed, dwFrameLimit, hNextFrameEvent, ctypes.byref(res))
        return res

    def NuiImageStreamGetNextFrame(self, hStream, dwMillisecondsToWait):
        res = ImageFrame()
        _NuiInstance._NuiImageStreamGetNextFrame(self, hStream, dwMillisecondsToWait, ctypes.byref(res))
        return res

    def NuiImageStreamReleaseFrame(self, hStream, pImageFrame):
        _NuiInstance._NuiImageStreamReleaseFrame(self, hStream, pImageFrame)

    def NuiImageGetColorPixelCoordinatesFromDepthPixel(self, eColorResolution, pcViewArea, lDepthX, lDepthY, usDepthValue):
        x, y = ctypes.c_long(), ctypes.c_long()
        _NuiInstance._NuiImageGetColorPixelCoordinatesFromDepthPixel(self, eColorResolution, pcViewArea, lDepthX, lDepthY, usDepthValue, ctypes.byref(x), ctypes.byref(y))
        return x.value, y.value

    def NuiCameraElevationSetAngle(self, lAngleDegrees):
        while 1:
            try:
                _NuiInstance._NuiCameraElevationSetAngle(self, lAngleDegrees)
                return
            except:
                pass

    def NuiCameraElevationGetAngle(self):
        res = ctypes.c_long()
        _NuiInstance._NuiCameraElevationGetAngle(self, ctypes.byref(res))
        return res.value

    def NuiSkeletonTrackingEnable(self, hNextFrameEvent = 0, dwFlags = 0):
        """Enables skeleton tracking.  Succeeds or raises an exception.

        hNextFrameEvent: A handle to an application-allocated, manual reset event that will be set whenever a new frame
        of skeleton data is available, and will be reset whenever the latest frame data is returned.
        This can be None.

        dwFlags: Flags that control skeleton tracking, as a bitwise-OR combination SkeletonTracking values.
        """
        _NuiInstance._NuiSkeletonTrackingEnable(self, hNextFrameEvent, dwFlags)

    def NuiSkeletonTrackingDisable(self):
        _NuiInstance._NuiSkeletonTrackingDisable()

    def NuiSkeletonGetNextFrame(self, dwMillisecondsToWait):
        frame = SkeletonFrame()
        _NuiInstance._NuiSkeletonGetNextFrame(self, dwMillisecondsToWait, ctypes.byref(frame))
        return frame

    def NuiTransformSmooth(self, pSkeletonFrame, pSmoothingParams):
        _NuiInstance._NuiTransformSmooth(self, pSkeletonFrame, pSmoothingParams)

    def GetUniqueDeviceName(self):        
        mem = ctypes.c_voidp()
        # Size is currently not used, and when we get the unique device name we need to free the memory.

        _NuiInstance._MSR_NuiGetPropsBlob(self, _PropsIndex.INDEX_UNIQUE_DEVICE_NAME, ctypes.byref(mem), None)
        res = ctypes.cast(mem, ctypes.c_wchar_p).value
        _SysFreeString(mem)
        return res

    def MSR_NuiGetPropsType(index):
        return _NuiInstance._MSR_NuiGetPropsType(index)


##***********************
## NUI enumeration function
##***********************

__NuiGetSensorCount = _NUIDLL.NuiGetSensorCount
__NuiGetSensorCount.argtypes = [ctypes.POINTER(ctypes.c_int)]
__NuiGetSensorCount.restype = ctypes.HRESULT

def _NuiGetSensorCount():
    count = ctypes.c_int()
    __NuiGetSensorCount(ctypes.byref(count))
    return count.value
    
__NuiCreateSensorByIndex = _NUIDLL.NuiCreateSensorByIndex
__NuiCreateSensorByIndex.argtypes = [ctypes.c_int, ctypes.POINTER(_NuiInstance)]
__NuiCreateSensorByIndex.restype = ctypes.HRESULT

def _NuiCreateSensorByIndex(index):
    inst = _NuiInstance()
    __NuiCreateSensorByIndex(index, ctypes.byref(inst))
    return inst



from PyQt5.QtWidgets import *
from PyQt5.QtGui import *

from pygments.formatter import Formatter
from pygments.lexers.python import PythonLexer


class Highlighter(QSyntaxHighlighter):

    style_name = "solarized-dark"

    def __init__(self, parent):
        QSyntaxHighlighter.__init__(self, parent)

        self.formatter = Formatter(style=self.style_name)
        self.lexer = PythonLexer()
        self.style = {}
        for token, style in self.formatter.style:
            char_format = QTextCharFormat()
            if style['color']:
                char_format.setForeground(QColor("#" + style['color']))
            if style['bgcolor']:
                char_format.setBackground(QColor("#" + style['bgcolor']))
            if style['bold']:
                char_format.setFontWeight(QFont.Bold)
            if style['italic']:
                char_format.setFontItalic(True)
            if style['underline']:
                char_format.setFontUnderline(True)

            char_format.setFontStyleHint(QFont.Monospace)
            self.style[token] = char_format

    def highlightBlock(self, text):
        position = self.currentBlock().position()
        length = self.currentBlock().length()
        text = str(self.document().toPlainText()[position:position+length])
        tokens = self.lexer.get_tokens(text)
        i = 0
        for token, text in tokens:
            self.setFormat(i, len(text), self.style[token])
            i += len(text)



from sklearn.ensemble import RandomForestClassifier
from sklearn.naive_bayes import BernoulliNB
from sklearn.naive_bayes import GaussianNB
from sklearn.tree import DecisionTreeClassifier

from cvlab.diagram.elements.base import *

from .ml import Trainable


class ScikitSimpleTrainable(Trainable):

    def __init__(self):
        super(ScikitSimpleTrainable, self).__init__()
        self.model = None

    def build_classifier(self):
        pass

    def get_attributes(self):
        return [Input("train data"),
                Input("responses")], \
               [Output("model"),
                Output("log")], \
               []

    def process_inputs(self, inputs, outputs, parameters):
        train_data = inputs["train data"].value
        responses = inputs["responses"].value
        sample_weights = self.get_sample_weights(responses)
        log = self.train_and_evaluate(train_data, responses, sample_weights)
        outputs["model"] = Data(self.model)
        outputs["log"] = Data(log)

    def train(self, train_data, responses, sample_weights):
        self.model = self.build_classifier()
        try:
            self.model.fit(train_data, responses, sample_weight=sample_weights)
        except TypeError:
            self.model.fit(train_data, responses)

    def predict(self, v):
        return self.model.predict(v)


class ScikitSimplePrediction(NormalElement):

    name = "Predict"
    comment = "General prediction for scikit classifiers"

    def __init__(self):
        super(ScikitSimplePrediction, self).__init__()
        self.model = None

    def get_attributes(self):
        return [Input("test data"),
                Input("model")], \
               [Output("responses")], \
               []

    def process_inputs(self, inputs, outputs, parameters):
        self.model = inputs["model"].value
        self.test_data = inputs["test data"].value
        predictions = []
        for v in self.test_data:
            result = self.model.predict(v)
            predictions.append(result)
        result = np.array(predictions)
        outputs["responses"] = Data(result)


class NaiveBayesBernoulliTrain(ScikitSimpleTrainable):
    name = "NaiveBayesBernoulliTrain"
    comment = "Bernoulli Naive Bayes classifier training"

    def build_classifier(self):
        return BernoulliNB()


class NaiveBayesGaussTrain(ScikitSimpleTrainable):
    name = "NaiveBayesGaussTrain"
    comment = "Gaussian Naive Bayes classifier training"

    def build_classifier(self):
        return GaussianNB()


class DecisionTreeTrain(ScikitSimpleTrainable):
    name = "DecisionTreeTrain"
    comment = "Decision Tree classifier training"

    def __init__(self):
        super(DecisionTreeTrain, self).__init__()
        self.model = None
        self.max_depth = None
        self.min_samples_split = 2
        self.min_samples_leaf = 1

    def build_classifier(self):
        return DecisionTreeClassifier(
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf
        )

    def get_attributes(self):
        return [Input("train data"),
                Input("responses")], \
               [Output("model"),
                Output("log")], \
               [
                   IntParameter("max_depth", value=0, min_=0),
                   IntParameter("min_samples_split", value=2, min_=0),
                   IntParameter("min_samples_leaf", value=1, min_=0)
               ]

    def process_inputs(self, inputs, outputs, parameters):
        train_data = inputs["train data"].value
        responses = inputs["responses"].value
        self.max_depth = parameters["max_depth"] if parameters["max_depth"] != 0 else None
        self.min_samples_split = parameters["min_samples_split"]
        self.min_samples_leaf = parameters["min_samples_leaf"]
        sample_weights = self.get_sample_weights(responses)
        log = self.train_and_evaluate(train_data, responses, sample_weights)
        outputs["model"] = Data(self.model)
        outputs["log"] = Data(log)

    def train(self, train_data, responses, sample_weights):
        self.model = self.build_classifier()
        self.model.fit(train_data, responses, sample_weight=sample_weights)


class RandomForestTrain(ScikitSimpleTrainable):
    name = "RandomForestTrain"
    comment = "Random Forest classifier training"

    def build_classifier(self):
        return RandomForestClassifier()


register_elements_auto(__name__, locals(), "Machine learning - scikit", 12)




from .base import *


class TextComment(NormalElement):
    name = 'Comment'
    comment = 'Element containing text editor. Double click to edit content. ' \
              'Editor supports text formatting with HTML tags'

    def get_attributes(self):
        return [], [], [CommentParameter(id="editor", value="Double click to edit...")]


register_elements_auto(__name__, locals(), "Diagram utils", 30)



#!/usr/bin/env python3

import sys
import os
from distutils.core import setup

from setuptools import find_packages
from glob import glob


data_files = glob("cvlab/images/*.*") + \
             glob("cvlab/styles/**/*.*", recursive=True)
data_files = list(map(lambda x:x[6:], data_files))

sample_files = glob("cvlab_samples/*.cvlab") + glob("cvlab_samples/images/*.*")
sample_files = list(map(lambda x:x[14:], sample_files))

if sys.version_info.major <= 2:
    raise Exception("Only python 3+ is supported!")

is_windows = os.name == 'nt'
is_linux = os.name == "posix"


requirements = [
    "numpy",
    "scipy",
    "pygments>=2",
    "matplotlib",
    "tinycss2",
]

try:
    import cv2
    if str(cv2.__version__) < "3":
        print("WARNING! OpenCV version 2.x detected. It is *strongly advised* to install OpenCV 3.x")
        print("Please visit: https://opencv.org/releases.html")
except ImportError:
    print("ERROR! OpenCV is required. Trying to use opencv-python package...")
    print("If it doesn't work, please visit and install OpenCV manually: https://opencv.org/releases.html")
    requirements.append("opencv-python")


try:
    import PyQt5
except ImportError:
    print("ERROR! PyQt5 is required. Trying to use PyPi package...")
    print("If it doesn't work, please install PyQt5 manually, e.g.:")
    print("    $ pip install --user pyqt5")
    print(" or $ sudo apt install python3-pyqt5")
    requirements.append("pyqt5")


__version__ = None
package_name = None
exec(compile(open('cvlab/version.py').read(), 'cvlab/version.py', 'exec'))


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

description = long_description.splitlines()[0].strip()

setup(
    name=package_name,
    version=__version__,
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Adam Brzeski, Jan Cychnerski',
    author_email='adam.m.brzeski@gmail.com, jan.cychnerski@gmail.com',
    url='https://github.com/cvlab-ai/cvlab',
    packages=find_packages(exclude=["diagrams","tools","samples","temp"]),
    package_data={"cvlab":data_files, "cvlab_samples":sample_files},
    entry_points={'gui_scripts': ['cvlab=cvlab.__main__:main']},
    license="AGPL-3.0+",
    python_requires='>=3.3',
    install_requires=requirements,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Code Generators",
    ],
)



from .base import *


class OpenCVMorphologyEx(NormalElement):
    name = "Morphological transform"
    comment = "Advanced morphological transform"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [
                    ComboboxParameter("operation", values=OrderedDict([
                        ("MORPH_OPEN", cv.MORPH_OPEN),
                        ("MORPH_CLOSE", cv.MORPH_CLOSE),
                        ("MORPH_GRADIENT", cv.MORPH_GRADIENT),
                        ("MORPH_TOPHAT ", cv.MORPH_TOPHAT),
                        ("MORPH_BLACKHAT", cv.MORPH_BLACKHAT)
                    ])),
                    ComboboxParameter("element type", values=OrderedDict([
                        ("MORPH_RECT", cv.MORPH_RECT),
                        ("MORPH_ELLIPSE", cv.MORPH_ELLIPSE),
                        ("MORPH_CROSS", cv.MORPH_CROSS)
                    ])),
                    IntParameter("element size", "Size of structuring element", 5, min_=1, max_=255, step=2),
                    IntParameter("iterations", "Number of iterations", 1, min_=0, max_=255)
               ]

    def process_inputs(self, inputs, outputs, parameters):
        image = np.copy(inputs["input"].value)
        operation = parameters["operation"]
        element_type = parameters["element type"]
        element_size = parameters["element size"]
        iterations = parameters["iterations"]

        element = cv.getStructuringElement(element_type, (element_size, element_size))
        image = cv.morphologyEx(image, operation, element, iterations=iterations,)
        outputs["output"] = Data(image)


class OpenCVDilate(NormalElement):
    name = "Dilate"
    comment = "Dilation morphological transform"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [
                    ComboboxParameter("element type", values=OrderedDict([
                        ("MORPH_RECT", cv.MORPH_RECT),
                        ("MORPH_ELLIPSE", cv.MORPH_ELLIPSE),
                        ("MORPH_CROSS", cv.MORPH_CROSS)
                    ])),
                    IntParameter("element size", "Size of structuring element", 5, min_=1, max_=255, step=2),
                    IntParameter("iterations", "Number of iterations", 1, min_=0, max_=255)
               ]

    def process_inputs(self, inputs, outputs, parameters):
        image = np.copy(inputs["input"].value)
        element_type = parameters["element type"]
        element_size = parameters["element size"]
        iterations = parameters["iterations"]

        element = cv.getStructuringElement(element_type, (element_size, element_size))
        image = cv.dilate(image, element, iterations=iterations)
        outputs["output"] = Data(image)


class OpenCVErode(NormalElement):
    name = "Erode"
    comment = "Erosion morphological transform"

    def get_attributes(self):
        return [Input("input")], \
               [Output("output")], \
               [
                    ComboboxParameter("element type", values=OrderedDict([
                        ("MORPH_RECT", cv.MORPH_RECT),
                        ("MORPH_ELLIPSE", cv.MORPH_ELLIPSE),
                        ("MORPH_CROSS", cv.MORPH_CROSS)
                    ])),
                    IntParameter("element size", "Size of structuring element", 5, min_=1, max_=255, step=2),
                    IntParameter("iterations", "Number of iterations", 1, min_=0, max_=255)
               ]

    def process_inputs(self, inputs, outputs, parameters):
        image = np.copy(inputs["input"].value)
        element_type = parameters["element type"]
        element_size = parameters["element size"]
        iterations = parameters["iterations"]

        element = cv.getStructuringElement(element_type, (element_size, element_size))
        image = cv.erode(image, element, iterations=iterations)
        outputs["output"] = Data(image)


register_elements_auto(__name__, locals(), "Filters", 5)
